# -*- coding: utf-8 -*-

import os
import tempfile
import unittest

from uval.tests.test_hdf5 import write_sample_h5_file
from uval.utils.hdf5_verification import *


class TestFormatMethods(unittest.TestCase):

    temp_h5_path = ""
    sample_h5 = None

    def __init__(self, *args, **kwargs):
        super(TestFormatMethods, self).__init__(*args, **kwargs)
        self.tempdir_object = tempfile.TemporaryDirectory()
        self.tempdir = self.tempdir_object.name
        self.temp_h5_path = os.path.join(self.tempdir, "test_format.h5")
        write_sample_h5_file(self.temp_h5_path)
        self.sample_h5 = UvalHdfFileInput(self.temp_h5_path)

    def __del__(self):
        if os.path.exists(self.temp_h5_path):
            os.unlink(self.temp_h5_path)

    def test_check_fields(self):
        with UvalHdfFileInput(self.temp_h5_path) as f:
            functions = [
                h5_check_file_meta_fields,
                h5_check_detection_fields,
                h5_check_groundtruth_fields,
                h5_check_volume_meta_fields,
            ]
            for func in functions:
                try:
                    func(f.h5)
                except ValueError:
                    self.fail(f"'{func}' raised ValueError unexpectedly!")

    def test_for_not_existing_fields(self):
        with UvalHdfFileInput(self.temp_h5_path) as f:
            self.assertRaises(ValueError, h5_check_volcache, f.h5)

    def test_verify_single_file(self):
        verify_single_hdf5_file(self.temp_h5_path)


if __name__ == "__main__":
    unittest.main()



import os
from datetime import datetime
from os import path

import randomname  # type: ignore
from fvcore.common.config import CfgNode as CN  # type: ignore

# -----------------------------------------------------------------------------
# Config definition
# -----------------------------------------------------------------------------
CPUS = os.cpu_count() if os.cpu_count() else 1
UVAL_DIR = path.abspath(path.join(path.dirname(__file__), "..", "..", ".."))
_C = CN()

_C.ENV = CN()
_C.ENV.CACHE_FOLDER = UVAL_DIR

_C.DATA = CN()
_C.DATA.PATH = path.join(UVAL_DIR, "data", "hdf5")
_C.DATA.IGNORE_MISSING_FILES = False
_C.DATA.MAX_THREADS = CPUS

_C.DATA_SPLIT = CN()
_C.DATA_SPLIT.YAML = path.join(UVAL_DIR, "data", "datasplit", "uval_ds.yaml")
_C.DATA_SPLIT.SUBSET = ["train", "test"]


_C.METRICS = CN()
# f-{factor] score
_C.METRICS.FACTOR = 1
_C.METRICS.IOU_THRESHOLD = 0.3
_C.METRICS.IOU_RANGE = (0,)
_C.METRICS.MAX_PROCESSES = CPUS
_C.METRICS.CONFIDENCE_THRESHOLD = 0.6
_C.METRICS.AP_METHOD = "EveryPointInterpolation"
tt = datetime.now()

_C.OUTPUT = CN()
_C.OUTPUT.PATH = path.abspath(path.join(UVAL_DIR, "output", tt.strftime("%Y%m%d%H%M%S")))
_C.OUTPUT.TITLE = randomname.get_name()
_C.OUTPUT.TEMPLATES_PATH = path.abspath(path.join(UVAL_DIR, "src", "uval", "templates"))
_C.OUTPUT.CONFIG_FILE = "config.yaml"
_C.OUTPUT.REPORT_FILE = "report.html"
_C.OUTPUT.METRICS_FILE = "metrics.pickle"
_C.OUTPUT.DATASET_OVERVIEW_FILE = "dataset_overview.csv"



from os import path  # type: ignore

import pytest  # type: ignore

from uval.config import default_argument_parser  # type: ignore
from uval.config.config_utils import setup_from_args  # type: ignore


@pytest.fixture
def setup():
    """
    Create configs and perform basic setups.
    """
    parser = default_argument_parser()
    args = parser.parse_args()
    cfg = setup_from_args(args)
    if isinstance(cfg, list):
        return cfg[0]
    else:
        return cfg


def test_config(setup):
    assert path.isdir(setup.ENV.CACHE_FOLDER)
    assert path.isdir(setup.DATA.PATH)
    assert type(setup.DATA.IGNORE_MISSING_FILES) == bool
    assert path.isfile(setup.DATA_SPLIT.YAML)
    assert type(setup.DATA_SPLIT.SUBSET) == list and len(setup.DATA_SPLIT.SUBSET) >= 1
    assert type(setup.METRICS.FACTOR) == int and setup.METRICS.FACTOR >= 1
    assert type(setup.METRICS.IOU_THRESHOLD) == float and 0.0 <= setup.METRICS.IOU_THRESHOLD <= 1.0
    assert type(setup.METRICS.CONFIDENCE_THRESHOLD) == float and 0.0 <= setup.METRICS.CONFIDENCE_THRESHOLD <= 1.0
    assert setup.METRICS.AP_METHOD == "EveryPointInterpolation" or setup.METRICS.AP_METHOD == "ElevenPointInterpolation"

    assert path.isdir(setup.OUTPUT.PATH)
    assert path.isdir(setup.OUTPUT.TEMPLATES_PATH)



# -*- coding: utf-8 -*-
"""
    We defined a 'short' form of label names.
    That means, if we have a bag with volume id
    BAGGAGE_20171205_081937_012345
    and a label called
    BAGGAGE_20171205_081937_012345_label_2,
    we can write the label as '%_label_2'.

    To escape a '%' as actual character, we use '%%' in the short form.

    This module provides functions to convert between the two forms.
"""

import re

_lonely_percent_regex = re.compile(r"(?<!%)%(?!%)")  # A '%' with no '%' before or after


def label_short_to_long(volume_id: str, short_label_id: str) -> str:
    """
        This function will restore the full label_id if it is a short form.
        For a definition of the short and long form, see this module's docstring.
    Args:
        volume_id: The ID which represents a image file containing the volume
        short_label_id: A shortened version of label id containing '%' character

    Returns:
        Extended label_id, in which '%' is replaced by the volume_id
    """
    if "%" not in short_label_id:
        raise ValueError(f"Given short id '{short_label_id}' must contain '%' character")

    return _lonely_percent_regex.sub(volume_id, short_label_id).replace("%%", "%")


def label_long_to_short(volume_id: str, long_label_id: str) -> str:
    """
        This function will shorten the full label_id if possible.
        For a definition of the short and long form, see this module's docstring.
    Args:
        volume_id: The ID which represents a image file containing the volume
        long_label_id: Full id of label with no '%' inside

    Returns:
        Shortened label_id, in which the volume_id is replaced by '%'
    """
    if volume_id not in long_label_id:
        raise ValueError(f"The long id {long_label_id} must contain the '{volume_id}' short id")

    return long_label_id.replace("%", "%%").replace(volume_id, "%")



import argparse
import os
from os import path

from uval.utils.log import logger


def default_argument_parser(epilog=None):
    """
    Create a parser with some common arguments used by uval users.
    Args:
        epilog (str): epilog passed to ArgumentParser describing the usage.
    Returns:
        argparse.ArgumentParser:
    """
    parser = argparse.ArgumentParser(
        epilog=epilog or """ Examples: None existent at the moment.""",
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )
    parser.add_argument(
        "--config-file", default="", metavar="FILE", help="path to config file or files, comma separated"
    )
    parser.add_argument(
        "opts",
        help="""
            Modify config options at the end of the command. For Yacs configs, use
            space-separated "PATH.KEY VALUE" pairs.
            For python-based LazyConfig, use "path.key=value".
        """.strip(),
        default=None,
        nargs=argparse.REMAINDER,
    )
    return parser


def get_cfg():
    """
    Get a copy of the default config.
    Returns:
        a Uval CfgNode instance.
    """
    from .defaults import _C

    return _C.clone()


def setup_from_args(args):
    """
    Create configs and perform basic setups.
    """
    cfgs = []
    config_files = args.config_file.split(",")
    if not config_files:
        cfg = get_cfg()
        cfg_process(cfg, args.opts)
        cfgs.append(cfg)

    for config_file in config_files:
        cfg = get_cfg()
        if path.isfile(config_file):
            cfg.merge_from_file(config_file)
        cfg_process(cfg, args.opts)
        cfgs.append(cfg)
    return cfgs


def cfg_process(cfg, opts):
    cfg.merge_from_list(opts)
    cfg.freeze()
    os.makedirs(cfg.OUTPUT.PATH, exist_ok=True)
    cfg_str = cfg.dump()
    with open(os.path.join(cfg.OUTPUT.PATH, cfg.OUTPUT.CONFIG_FILE), "w") as f:
        f.write(cfg_str)
    logger.info(f"config file saved to {os.path.join(cfg.OUTPUT.PATH, cfg.OUTPUT.CONFIG_FILE)}.")



# -*- coding: utf-8 -*-
"""A StageData object acts like a data container that manages results generated by stages and read by subsequent
stages.
"""

from uval.stages.stage import uval_stage  # type: ignore
from uval.stages.stage_data import Hdf5FilesData, SupportedDatasetSpecificationData  # type: ignore
from uval.utils.log import logger  # type: ignore


@uval_stage
def create_supported_dataset_from_intersection(gt_files: Hdf5FilesData, det_files: Hdf5FilesData):
    """Given a list of ground truth and detection files, try to join them given their volume ids."""

    # Let GT be a set of ground truth volume ids, DET be a set of detection volume ids
    # Determine GT & DET set, and the size of the GT-DET and DET-GT sets
    gt_ids = set(gt_files.table["volume_id"])
    det_ids = set(det_files.table["volume_id"])

    set_intersection = gt_ids & det_ids
    set_gt_minus_det = gt_ids - det_ids
    set_det_minus_gt = det_ids - gt_ids

    if len(set_intersection) == 0:
        logger.error("Did not find any matching files between groundtruth and detections")
        return SupportedDatasetSpecificationData()  # Return empty set

    # This is important feedback for the user to see if this matches their expectations
    logger.info(
        f"Dataset intersection statistics:\n"
        f"  GT & DET = {len(set_intersection)}\n"
        f"  GT - DET = {len(set_gt_minus_det)}\n"
        f"  DET - GT = {len(set_det_minus_gt)}\n"
    )

    result = SupportedDatasetSpecificationData()

    # Now create the supported dataset from the intersection set
    idx = 0
    for vol_id in sorted(list(set_intersection)):
        cur_gt_file = gt_files.table.loc[vol_id]
        cur_det_file = det_files.table.loc[vol_id]

        # Now iterate over every ground truth label
        for gt_label, gt in cur_gt_file["hdf5_meta"]["groundtruth"].items():
            result.table.loc[idx] = {
                "label_id": gt_label,
                "volume_id": vol_id,
                "is_negative": False,  # For now, no SOC data used here
                "target_id": gt["target_id"],
                "subset": None,  # Since this is not based on a data split, we don't have a subset
                "class_name": gt["class_name"],
                "hdf5_detection": cur_det_file,
                "hdf5_groundtruth": cur_gt_file,
            }
            idx += 1

    # import pandas
    # pandas.set_option("display.max_rows", None, "display.max_columns", None)
    # pandas.set_option("display.max_colwidth", 200, "display.width", 5000)
    # print(result.table)
    # result.table.to_html("test.htm")

    return result


@uval_stage
def support_dataset_with_file_paths(gt_files: Hdf5FilesData, det_files: Hdf5FilesData, soc_files: Hdf5FilesData):
    """Given a dataset and sets of Hdf5 files found on disk, try to combine everything to
    create a supported dataset."""

    # Let GT be a set of ground truth volume ids, DET be a set of detection volume ids
    # Determine GT & DET set, and the size of the GT-DET and DET-GT sets
    gt_ids = set(gt_files.table["volume_id"])
    det_ids = set(det_files.table["volume_id"])
    soc_ids = set(soc_files.table["volume_id"])

    set_union = gt_ids | det_ids | soc_ids
    set_intersection = gt_ids & det_ids
    set_gt_minus_det = gt_ids - det_ids
    set_det_minus_gt = det_ids - gt_ids

    if len(set_union) == 0:
        logger.error("Did not find any matching files between groundtruth and detections")
        return SupportedDatasetSpecificationData()  # Return empty set

    # This is important feedback for the user to see if this matches their expectations
    logger.info(
        f"Dataset union statistics:\n"
        f"  GT & DET = {len(set_intersection)}\n"
        f"  GT - DET = {len(set_gt_minus_det)}\n"
        f"  DET - GT = {len(set_det_minus_gt)}\n"
        f"  DET + GT + SOC  = {len(set_union)}\n"
    )

    result = SupportedDatasetSpecificationData()

    # Now create the supported dataset from the intersection set
    idx = 0
    for vol_id in sorted(list(set_intersection)):
        # Now iterate over every ground truth label

        result.table.loc[idx] = {
            "volume_id": vol_id,
            "is_negative": False,  # For now, no SOC data used here
            "hdf5_detection": det_files.table.loc[vol_id]["hdf5_meta"]["detections"],
            "hdf5_groundtruth": gt_files.table.loc[vol_id]["hdf5_meta"]["groundtruth"],
        }
        idx += 1
    for vol_id in sorted(list(soc_ids)):
        # Now iterate over every ground truth label
        result.table.loc[idx] = {
            "volume_id": vol_id,
            "is_negative": True,
            "hdf5_detection": soc_files.table.loc[vol_id]["hdf5_meta"]["detections"],
            "hdf5_groundtruth": None,
        }
        idx += 1
    return result



# -*- coding: utf-8 -*-
"""A StageData object acts like a data container that manages results generated by stages and read by subsequent
stages.
"""

from os import path

from uval.stages.stage_data import DatasetSpecificationData
from uval.utils.label_naming import label_short_to_long
from uval.utils.log import logger
from uval.utils.yaml_io import load_yaml_data


# @uval_stage
def load_datasplit(file_path: str, subsets=None, output=None) -> DatasetSpecificationData:
    """This stage will load a data split file in YAML format and keep the id lists for
    train, val and test set.

    Args:
        file_path (str): path to the yaml file.
        subsets ([str], optional): which subsets of the yaml files are to be loaded,
        if set to None, all subsets are used. Defaults to None.
        output: The output file to be generated, if needed

    Returns:
        DatasetSpecificationData: returns the dataset.
    """

    # Set up result data store
    stage_results = DatasetSpecificationData()
    logger.debug("Reading YAML file started")
    data_split_dict = load_yaml_data(file_path)
    logger.debug("Reading YAML file finished.")

    if not data_split_dict or not data_split_dict.get("split"):
        logger.error("No data could be read from data split file '{}'".format(file_path))
        return DatasetSpecificationData()
    available_subsets = list()
    if data_split_dict.get("split"):
        available_subsets = [subset_name for subset_name in data_split_dict.get("split").keys()]  # type: ignore
    if subsets is None:
        subsets = available_subsets
    else:
        for subset in subsets:
            if subset not in available_subsets:
                raise ValueError(f"The requested split subset {subset} not found in the YAML file.")
    logger.debug(subsets)

    # frame_columns = ["volume_id", "label_id", "is_negative", "subset", "class_name"]
    dic = dict()

    for subset_name, subset_data in data_split_dict.get("split").items():  # type: ignore
        if subset_name not in subsets:
            continue
        for class_name, image_list in subset_data.items():
            is_negative = True if class_name == "negative" else False
            logger.debug("Reading list of '{}' images for split {}/{}".format(len(image_list), subset_name, class_name))
            for volume_id, labels in image_list.items():
                labels_list = labels if isinstance(labels, list) else [labels]
                for label_id in labels_list:
                    try:
                        item_key = label_short_to_long(volume_id, label_id) if label_id else volume_id
                        dic[item_key] = {
                            "volume_id": volume_id,
                            "label_id": label_id,
                            "is_negative": is_negative,
                            "subset": subset_name,
                            "class_name": class_name,
                        }
                    except KeyError:
                        logger.warning(f"Unable to find proper keys in the YAML file: '{file_path}'")

    stage_results.from_dict_as_rows(dic)
    if output:
        if output.DATASET_OVERVIEW_FILE.split(".")[-1] == "html":
            stage_results.to_html(path.join(output.PATH, output.DATASET_OVERVIEW_FILE))
        elif output.DATASET_OVERVIEW_FILE.split(".")[-1] == "csv":
            stage_results.to_csv(path.join(output.PATH, output.DATASET_OVERVIEW_FILE))
        else:
            raise NotImplementedError("Other output formats rather than 'html' and 'csv' are not implemented yet!")
        logger.info(f"dataset overview saved to {path.join(output.PATH, output.DATASET_OVERVIEW_FILE)}.")

    return stage_results



# -*- coding: utf-8 -*-
"""This module provides stages that can compute metrics like overlaps between
groundtruth and detections, or simply count detections per bag.
"""
import math
import os
import pickle
import sys
from multiprocessing import Pool
from typing import Any, Dict, List, Tuple, Union

import jinja2  # type: ignore
import matplotlib  # type: ignore
import matplotlib.pyplot as plt  # type: ignore
import numpy as np
import pandas as pd  # type: ignore
from rich.progress import track

from uval.stages.stage import uval_stage  # type: ignore
from uval.stages.stage_data import SupportedDatasetSpecificationData
from uval.utils.log import logger

matplotlib.use("Agg")


class DetectionEntry:
    """Any detection entry wether it is from GT or detections will be converted in this form."""

    def __init__(
        self, volume_id: str, class_name: str, confidence_score: float, roi_start: Tuple[int], roi_shape: Tuple[int]
    ):
        self.volume_id = volume_id
        self.class_name = class_name
        self.confidence_score = confidence_score
        self.roi_start = roi_start
        self.roi_shape = roi_shape


class Metrics:
    """This class implements all the evaluation metrics."""

    def __init__(self, dataset: SupportedDatasetSpecificationData, metrics_settings, output_settings):
        self.dataset = dataset

        self.data = None
        self.iou_threshold = metrics_settings.IOU_THRESHOLD

        self.template_file = "template.html"
        if len(metrics_settings.IOU_RANGE) == 3:
            a, b, c = metrics_settings.IOU_RANGE
            self.iou_range = np.linspace(a, b, int(np.round((b - a) / c)) + 1, endpoint=True).tolist()
            self.iou_range = [round(iou, 2) for iou in self.iou_range]
            self.template_file = "template_range.html"
        else:
            self.iou_range = None
        self.confidence_threshold = metrics_settings.CONFIDENCE_THRESHOLD
        self.output_path = output_settings.PATH

        self.title = output_settings.TITLE or self.output_path.split("/")[-1]

        self.report_file = output_settings.REPORT_FILE
        self.metrics_file = output_settings.METRICS_FILE
        self.max_workers = metrics_settings.MAX_PROCESSES
        self.templates_path = output_settings.TEMPLATES_PATH
        self.ap_method = metrics_settings.AP_METHOD
        self.factor = metrics_settings.FACTOR

    def evaluate(self):
        metrics_output = dict()
        basics = self.basic_metric(iou_threshold=self.iou_threshold, confidence_threshold=self.confidence_threshold)
        ap_metrics = self.get_average_precision(basics, method=self.ap_method)
        fscore_metrics = self.get_fscore(ap_metrics)
        metrics_output["title"] = self.title
        metrics_output["iou_threshold"] = self.iou_threshold
        metrics_output["single_threshold"] = fscore_metrics
        metrics_output["confidence_threshold"] = self.confidence_threshold
        if self.iou_range:
            aps, rs, ars, map, mar = self.evaluate_range(iou_range=self.iou_range)
            metrics_output["AP"] = aps
            metrics_output["rs"] = rs
            metrics_output["ars"] = ars
            metrics_output["map"] = map
            metrics_output["mar"] = mar
            metrics_output["iou_range"] = self.iou_range
        with open(os.path.join(self.output_path, self.metrics_file), "wb") as f:
            pickle.dump(metrics_output, f)
        logger.info(f"metrics saved to {os.path.join(self.output_path, self.metrics_file)}.")
        self.plot_roc_curves(fscore_metrics)
        self.plot_precision_recall_curve(fscore_metrics)
        self.generate_report(metrics_output)
        return metrics_output

    def worker(self, iou_threshold):
        basics = self.basic_metric(iou_threshold=iou_threshold, confidence_threshold=0.1)

        rs = [result["Single_Recall"] for result in basics]
        output_metrics = self.get_average_precision(basics)
        aps = [result["AP"] for result in output_metrics]
        map = sum(aps) / len(aps)
        classes = [basic["Class"] for basic in basics]
        return rs, aps, map, classes

    def evaluate_range(self, iou_range: List[float], confidence_threshold: float = None):
        # if not iou_range:
        #    iou_range = self.iou_range
        if not confidence_threshold:
            confidence_threshold = self.confidence_threshold

        aps = dict()
        rs: Dict[float, List[float]] = dict()
        map = dict()
        with Pool(self.max_workers) as pool:

            result = pool.map(self.worker, iou_range)

        for iou, res in zip(iou_range, result):
            aps[iou] = res[0]
            rs[iou] = res[1]
            map[iou] = res[2]
            classes = res[3]

        assert len(aps) == len(iou_range)
        ars = self.get_average_recall(rs, iou_range)
        self.plot_recall_iou_curve(rs, iou_range, classes)

        mar = sum(ars) / len(ars)
        return aps, rs, ars, map, mar

    def plot_recall_iou_curve(
        self, recalls: Dict[float, List[float]], iou_thresholds: List[float], classes: List[str], show_ar: bool = True
    ) -> None:
        """Plot the Recall x IOU curve for a given class.

        Args:
            recalls (Dict[float:list]): keys are iou thresholds. value is a list of recall for each class.
            iou_thresholds: list of iou thresholds.
            classes: list containing names of all classes.
        """
        # Each result represents a class
        for index, class_id in enumerate(classes):
            recall_vector = [recalls[iou][index] for iou in iou_thresholds]

            plt.close()

            plt.plot(iou_thresholds, recall_vector, label=f"Confidence:{self.confidence_threshold}")
            plt.xlabel("IOU")
            plt.ylabel("Recall")
            if show_ar:

                ap_str = "{0:.2f}%".format(sum(recall_vector) / len(recall_vector) * 100)
                plt.title("Recall x IOU curve \nClass: %s, AR: %s" % (str(class_id), ap_str))
            else:
                plt.title("Recall x IOU curve \nClass: %s" % str(class_id))
            plt.legend(shadow=True)
            plt.grid()
            plt.xlim([0.0, 1.0])
            plt.ylim([0.0, 1.0])
            plt.savefig(os.path.join(self.output_path, class_id + "_recall_iou.png"))
        plt.close()

    @property
    def templates_path(self):
        return self._templates_path

    @templates_path.setter
    def templates_path(self, path):
        os.makedirs(os.path.abspath(path), exist_ok=True)
        self._templates_path = os.path.abspath(path)

    @property
    def output_path(self):
        return self._output_path

    @output_path.setter
    def output_path(self, path):
        os.makedirs(os.path.abspath(path), exist_ok=True)
        self._output_path = os.path.abspath(path)

    @property
    def iou_threshold(self):
        return self._iou_threshold

    @iou_threshold.setter
    def iou_threshold(self, value):
        if value < 0 or value > 1:
            raise ValueError
        self._iou_threshold = value

    @property
    def confidence_threshold(self):
        return self._confidence_threshold

    @confidence_threshold.setter
    def confidence_threshold(self, value):
        if value < 0 or value > 1:
            raise ValueError
        self._confidence_threshold = value

    @uval_stage
    def data_preparations(self):

        ground_truths = []

        detections = []
        # Get all classes
        classes = []
        det_classes = []
        volumes_soc = {}
        for row_num in track(range(len(self.dataset.table)), "Preparing..."):
            row = self.dataset.table.loc[row_num]

            volume_id = row["volume_id"]
            if volume_id not in volumes_soc:
                volumes_soc[volume_id] = 0

            det = row["hdf5_detection"]
            gt = row["hdf5_groundtruth"]
            if gt:
                volumes_soc.pop(volume_id, None)
                for gt_item in gt.values():
                    class_name = gt_item["class_name"]
                    ground_truths.append(
                        DetectionEntry(volume_id, class_name, 1.0, gt_item["roi_start"], gt_item["roi_shape"])
                    )
                    if class_name not in classes:
                        classes.append(class_name)
            # Loop through all bounding boxes and separate them into GTs and
            # detections

            for det_item in det:
                # if det_item["class_name"]=="bluntobject":
                #    class_name = "bluntobjects"
                if det_item["class_name"] == "iCMORE":
                    class_name = "grenade"
                else:
                    class_name = det_item["class_name"]

                det_classes.append(class_name)
                detections.append(
                    DetectionEntry(
                        volume_id, class_name, det_item["score"], det_item["roi_start"], det_item["roi_shape"]
                    )
                )
        logger.info(f"detected classes are:{set(det_classes)}")
        logger.info(f"ground truth classes are:{set(classes)}")
        total_negative = len(volumes_soc)
        return classes, volumes_soc, ground_truths, detections, total_negative

    @uval_stage
    def basic_metric(self, iou_threshold: float = None, confidence_threshold: float = None) -> List[dict]:
        """Get the TP, FP, Precision and recall.

        Args:
            iou_threshold (float, optional): Threshold for IOU. Defaults to None.
            confidence_threshold (float, optional): Threshold for confidence. Defaults to None.

        Returns:
            List[dict]: A list of dictionaries. Each dictionary contains information and
            metrics of each class.
        """

        if not self.data:
            self.data = self.data_preparations()
        classes, volumes_soc, ground_truths, detections, total_negative = self.data
        if iou_threshold is None:
            iou_threshold = self.iou_threshold
        if confidence_threshold is None:
            confidence_threshold = self.confidence_threshold
        ret = []
        # Loop through by classes
        for c in classes:
            volumes_negative_current = {v: 0 for v in volumes_soc}

            # Get only detection of class c
            dects = [d for d in detections if d.class_name == c]
            # dects_image_level = [d.volume_id for d in detections if d.class_name == c]
            # Get only ground truths of class c, use filename as key
            gts: Dict[Any, Any] = {}
            nneg = 0.0
            npos = 0.0
            for g in ground_truths:
                if g.class_name == c:
                    # volumes that do not contain class c
                    npos += 1.0
                    gts[g.volume_id] = gts.get(g.volume_id, []) + [g]
                else:
                    nneg += 1.0

            # sort detections by decreasing confidence
            dects = sorted(dects, key=lambda detected_entry: detected_entry.confidence_score, reverse=True)
            total_tp = 0
            total_fp = 0
            fp_image_level = np.zeros(len(dects))

            tp = np.zeros(len(dects))
            fp = np.zeros(len(dects))
            tp_soft = 0
            fp_soft = 0
            # create dictionary with amount of gts for each image
            det = {key: np.zeros(len(gts[key])) for key in gts}

            # Loop through detections
            single_recall = 0.0
            for d, dect in enumerate(dects):
                if dect.confidence_score < confidence_threshold:
                    single_recall = float(np.sum(tp)) / npos
                if dect.volume_id in volumes_negative_current:
                    fp_image_level[d] = 1
                    volumes_negative_current.pop(dect.volume_id, None)

                # Find ground truth image
                gt = gts[dect.volume_id] if dect.volume_id in gts else []
                iou_max = sys.float_info.min
                for j, gtj in enumerate(gt):
                    iou = Metrics.iou(dect.roi_start, dect.roi_shape, gtj.roi_start, gtj.roi_shape)
                    if iou > iou_max:
                        iou_max = iou
                        jmax = j
                # Assign detection as true positive/don't care/false positive
                if iou_max >= iou_threshold:
                    if det[dect.volume_id][jmax] == 0:
                        tp[d] = 1  # count as true positive
                        total_tp += 1 if dect.confidence_score > confidence_threshold else 0
                        tp_soft += dect.confidence_score if dect.confidence_score > confidence_threshold else 0
                        det[dect.volume_id][jmax] = 1  # flag as already 'seen'
                    else:
                        fp[d] = 1  # count as false positive
                        total_fp += 1 if dect.confidence_score > confidence_threshold else 0
                        fp_soft += dect.confidence_score if dect.confidence_score > confidence_threshold else 0
                # - A detected "cat" is overlaped with a GT "cat" with IOU >= iou_threshold.
                else:
                    fp[d] = 1  # count as false positive
                    total_fp += 1 if dect.confidence_score > confidence_threshold else 0
                    fp_soft += dect.confidence_score if dect.confidence_score > confidence_threshold else 0
            # compute precision, recall and average precision
            acc_fp = np.cumsum(fp)
            acc_tp = np.cumsum(tp)
            acc_fp_image_level = np.cumsum(fp_image_level)
            fpr = acc_fp_image_level / total_negative
            rec = acc_tp / npos

            prec = np.divide(acc_tp, (acc_fp + acc_tp))

            # Depending on the method, call the right implementation
            # add class result in the dictionary to be returned
            r = {
                "Class": c,
                "precision": prec,
                "recall": rec,
                "Single_Recall": single_recall,
                "Total Positives": npos,
                "Total Negatives": nneg,
                "Total TP": total_tp,
                "Total FP": total_fp,
                "Total FN": npos - total_tp,
                "Total TP soft": tp_soft,
                "Total FP soft": fp_soft,
                "Total FN soft": npos - tp_soft,
                "fpr": fpr,
            }
            ret.append(r)

        return ret

    def get_pascal_voc2012_metric(self, confidence_threshold=None) -> list:
        basics = self.basic_metric(iou_threshold=0.5, confidence_threshold=confidence_threshold)
        return self.get_average_precision(basics, method="EveryPointInterpolation")

    def get_pascal_voc2007_metric(self, confidence_threshold=None) -> list:
        basics = self.basic_metric(iou_threshold=0.5, confidence_threshold=confidence_threshold)
        return self.get_average_precision(basics, method="ElevenPointInterpolation")

    @uval_stage
    def get_average_recall(self, recalls: Dict[float, list], iou_range: List) -> List:
        class_num = len(recalls[iou_range[0]])
        average_recalls = [0] * class_num
        for c in range(class_num):
            area = 0
            for idx in range(len(iou_range) - 1):
                iou_step = iou_range[idx + 1] - iou_range[idx]
                area += iou_step * (
                    recalls[iou_range[idx]][c] + 0.5 * (recalls[iou_range[idx]][c] - recalls[iou_range[idx + 1]][c])
                )
            average_recalls[c] = area
        return average_recalls

    @uval_stage
    def get_average_precision(self, basic_metrics: List[dict], method: str = None) -> List[dict]:
        """Get the average precision. This will be used in multiple other metrics such as
        COCO or pascal voc.

        Args:
            basic_metrics (List[dict]): [description]
            method (str, optional): choice between precise (EveryPointInterpolation or None)
            or estimation (ElevenPointInterpolation). Defaults to None.

        Returns:
            List[dict]: adds ap to the each class of the output dictionaries.
        """

        if method is None:
            method = self.ap_method
        ret = []
        for err in basic_metrics:
            # Depending on the method, call the right implementation
            if method == "EveryPointInterpolation":
                [ap, mpre, mrec] = Metrics.calculate_average_precision(err["recall"], err["precision"])
            else:
                [ap, mpre, mrec] = Metrics.eleven_point_interpolated_ap(err["recall"], err["precision"])
            # add class result in the dictionary to be returned
            r = {key: value for key, value in err.items()}
            r["AP"] = ap
            r["interpolated precision"] = mpre
            r["interpolated recall"] = mrec

            ret.append(r)
        return ret

    @uval_stage
    def get_fscore(self, basic_metrics: List[dict]) -> List[dict]:
        """Get the f score metrics.

        Args:
            basic_metrics (List[dict]): output of basic_metric method.
            needs to be called before this method.

        Returns:
            List[dict]: adds dict['F score'] and dict['F score soft'] to the inputs.
        """

        ret = []
        for err in basic_metrics:
            fp = err["Total FP"]
            fn = err["Total FN"]
            tp = err["Total TP"]
            f_score = (1 + self.factor**2) * tp / ((1 + self.factor**2) * tp + (self.factor**2) * fn + fp)

            fp_soft = err["Total FP soft"]
            fn_soft = err["Total FN soft"]
            tp_soft = err["Total TP soft"]
            f_score_soft = (
                (1 + self.factor**2)
                * tp_soft
                / ((1 + self.factor**2) * tp_soft + (self.factor**2) * fn_soft + fp_soft)
            )
            # add class result in the dictionary to be returned
            r = {key: value for key, value in err.items()}
            r["F score"] = f_score
            r["F score soft"] = f_score_soft
            ret.append(r)
        return ret

    @uval_stage
    def generate_report(self, results_cluttered: dict) -> None:
        # Sample DataFrame
        range_results = dict(results_cluttered)
        single_results = range_results.pop("single_threshold")

        def func(row):

            highlight = "background-color: darkorange;"
            default = ""

            return [default] * (len(row) - 1) + [highlight]

        classes = []
        for idx, res in enumerate(single_results):
            res.pop("precision")
            res.pop("recall")
            res.pop("fpr")
            res.pop("interpolated precision")
            res.pop("interpolated recall")
            # res["ar"]=range_results["ars"][idx]
            classes.append(res["Class"])
            res.pop("Class")
        # for key, value in kwargs.items():
        high_level = dict()
        range_results.pop("ap", None)
        rs = range_results.pop("rs", None)
        range_results.pop("ars", None)
        range_results.pop("iou_range", None)
        high_level["mar"] = range_results.pop("mar", None)

        cell_hover = {  # for row hover use <tr> instead of <td>
            "selector": "td:hover",
            "props": [("background-color", "#ffffb3")],
        }
        row_hover = {  # for row hover use <tr> instead of <td>
            "selector": "tr:hover",
            "props": [("background-color", "#ffffb3")],
        }
        sorted_idx = [i[0] for i in sorted(enumerate(classes), key=lambda x: x[1])]
        single_results_sorted = [single_results[i] for i in sorted_idx]
        df = pd.DataFrame(single_results_sorted, index=pd.Index(sorted(classes)))

        styler = (
            df.style.set_caption(f"Calculated metrics for iou:{self.iou_threshold}")
            # .set_precision(2)
            .format(precision=2).set_table_styles([row_hover])
        )
        env = jinja2.Environment(loader=jinja2.FileSystemLoader(searchpath=self.templates_path))
        template = env.get_template(self.template_file)
        total_images = []

        for cls in sorted(classes):
            img_names = []
            img_names.append("." + "/" + cls + "_roc.png")
            img_names.append("." + "/" + cls + "_precision_recall.png")
            if self.iou_range:
                img_names.append("." + "/" + cls + "_recall_iou.png")
            total_images.append(img_names)

        if self.iou_range:
            high_level["map"] = sum(range_results["map"].values()) / len(range_results["map"])
            range_results["map"]["Total"] = high_level["map"]
            df2 = pd.DataFrame(
                [range_results["map"].values()], index=pd.Index(["map"]), columns=range_results["map"].keys()
            )
            rs_sorted = {key: [val[i] for i in sorted_idx] for key, val in rs.items()}
            df_rs = pd.DataFrame(rs_sorted, index=pd.Index(sorted(classes)))
            styler_rs = (
                df_rs.style.set_caption("Recall values for all classes and all IOU thresholds")
                .format(precision=2)
                .set_table_styles([row_hover])
            )

            styler2 = (
                df2.style.set_caption("Mean average precision for various IOU levels.")
                .format(precision=2)
                .set_table_styles([cell_hover])
                .apply(func, subset=["Total"], axis=1)
            )

            # Template handling

            html = template.render(
                range_table=styler2.to_html(),
                single_table=styler.to_html(),
                rs_table=styler_rs.to_html(),
                total_images=total_images,
                mar=round(high_level["mar"], 2),
                map=round(high_level["map"], 2),
                title=self.title,
            )
        else:
            html = template.render(single_table=styler.to_html(), total_images=total_images, title=self.title)
            # Template handling

        # Write the HTML file
        with open(os.path.join(self.output_path, self.report_file), "w") as f:
            f.write(html)
        logger.info(f"Report saved to {os.path.join(self.output_path, self.report_file)}.")

    @uval_stage
    def plot_precision_recall_curve(
        self,
        pascal_voc_metrics: List[dict],
        show_ap: bool = True,
        show_interpolated_precision: bool = True,
        show_graphic: bool = False,
    ) -> None:
        """Plot the Precision x Recall curve for a given class.

        Args:
            pascal_voc_metrics (List[dict]): Output of some pascal voc metric. needs to be
            called before this method.
            show_ap (bool, optional): if True, the average precision value will be shown
            in the title of the graph. Defaults to False.
            show_interpolated_precision (bool, optional): if True, it will show in the plot
            the interpolated precision. Defaults to False.
            show_graphic (bool, optional): if True, the plot will be shown. Defaults to False.

        Raises:
            IOError: [description]
        """

        result = None
        # Each result represents a class
        for result in pascal_voc_metrics:
            if result is None:
                raise IOError("Error: No data for a class was found.")

            class_id = result["Class"]
            precision = result["precision"]
            recall = result["recall"]
            average_precision = result["AP"]
            mpre = result["interpolated precision"]
            mrec = result["interpolated recall"]

            plt.close()
            if show_interpolated_precision:
                if self.ap_method == "EveryPointInterpolation":
                    plt.plot(mrec, mpre, "--r", label="Interpolated precision (every point)")
                elif self.ap_method == "ElevenPointInterpolation":
                    # Uncomment the line below if you want to plot the area
                    # plt.plot(mrec, mpre, 'or', label='11-point interpolated precision')
                    # Remove duplicates, getting only the highest precision of
                    # each recall value
                    nrec = []
                    nprec = []
                    for idx in range(len(mrec)):
                        r = mrec[idx]
                        if r not in nrec:
                            idx_eq = np.argwhere(mrec == r)
                            nrec.append(r)
                            nprec.append(max([mpre[int(idx)] for idx in idx_eq]))
                    plt.plot(nrec, nprec, "or", label="11-point interpolated precision")
                else:
                    raise NotImplementedError(
                        "plot_precision_recall_curve() without show_interpolated_precision is not implemented yet!"
                    )
            plt.plot(recall, precision, label=f"Precision for IOU:{self.iou_threshold}")
            plt.xlabel("recall")
            plt.ylabel("precision")
            if show_ap:
                ap_str = "{0:.2f}%".format(average_precision * 100)
                plt.title("Precision x Recall curve \nClass: %s, AP: %s" % (str(class_id), ap_str))
            else:
                plt.title("Precision x Recall curve \nClass: %s" % str(class_id))
            plt.legend(shadow=True)
            plt.grid()

            plt.xlim([0.0, 1.0])
            plt.ylim([0.0, 1.0])
            if self.output_path is not None:
                plt.savefig(os.path.join(self.output_path, class_id + "_precision_recall.png"))

            if show_graphic is True:
                plt.show(block=False)
                plt.pause(0.05)
        plt.close()

    @staticmethod
    def calculate_average_precision(rec: List[float], prec: List[float]) -> List[Any]:
        assert len(rec) == len(prec)
        mrec = [0.0] + list(rec) + [1.0]
        mpre = [1.0] + list(prec) + [0.0]
        for i in range(len(mpre) - 1, 0, -1):
            mpre[i - 1] = max(mpre[i - 1], mpre[i])
        ii = []
        for i in range(len(mrec) - 1):
            if mrec[1 + i] != mrec[i]:
                ii.append(i + 1)
        ap: Union[float, Any] = 0.0
        for i in ii:
            ap = ap + np.sum((mrec[i] - mrec[i - 1]) * mpre[i])
        # return [ap, mpre[1:len(mpre)-1], mrec[1:len(mpre)-1], ii]
        return [ap, mpre[0 : len(mpre) - 1], mrec[0 : len(mpre) - 1]]

    @staticmethod
    # 11-point interpolated average precision
    def eleven_point_interpolated_ap(rec: List[float], prec: List[float]) -> List[Any]:
        mrec = list(rec)
        mpre = list(prec)
        recall_values_np = np.linspace(0, 1, 11)
        recall_values = list(recall_values_np[::-1])
        rho_interp = []
        recall_valid = []
        # For each recall_values (0, 0.1, 0.2, ... , 1)
        for r in recall_values:
            # Obtain all recall values higher or equal than r
            arg_greater_recalls = np.argwhere(mrec[:] >= r)
            pmax = 0.0
            # If there are recalls above r
            if arg_greater_recalls.size != 0:
                pmax = max(mpre[int(arg_greater_recalls.min()) :])
            recall_valid.append(r)
            rho_interp.append(pmax)
        # By definition ap = sum(max(precision whose recall is above r))/11
        ap = sum(rho_interp) / 11
        # Generating values for the plot
        rvals = [recall_valid[0]] + list(recall_valid) + [0.0]
        pvals = [0.0] + list(rho_interp) + [0.0]
        # rho_interp = rho_interp[::-1]
        cc = []
        for i in range(len(rvals)):
            p = (rvals[i], pvals[i - 1])
            if p not in cc:
                cc.append(p)
            p = (rvals[i], pvals[i])
            if p not in cc:
                cc.append(p)
        recall_values_out = [i[0] for i in cc]
        rho_interp = [i[1] for i in cc]
        return [ap, rho_interp, recall_values_out]

    @staticmethod
    def iou(start_a: List[float], shape_a: List[float], start_b: List[float], shape_b: List[float]) -> float:
        """Calculates the intersection over union of the two cubes A and B.

        Args:
            start_a (List[float]): bottom left corner of the cube A.
            shape_a (List[float]): size of each dimension in the cube A.
            start_b (List[float]): bottom left corner of the cube B.
            shape_b (List[float]): size of each dimension in the cube B.

        Returns:
            float: 3D IOU of these cubes.
        """

        if (
            np.any(np.array(start_a) < 0)
            or np.any(np.array(start_b) < 0)
            or np.any(np.array(shape_a) < 0)
            or np.any(np.array(shape_b) < 0)
        ):
            logger.warning(f"bounding box coordinates are negative!{start_a}{shape_a}{start_b}{shape_b}")
            return 0

        if Metrics._boxes_intersect(start_a, shape_a, start_b, shape_b) is False:
            return 0
        inter_area = Metrics._get_intersection_area(start_a, shape_a, start_b, shape_b)
        union = Metrics._get_union_areas(start_a, shape_a, start_b, shape_b)
        # intersection over union
        iou = inter_area / union
        assert iou >= 0
        return iou

    @staticmethod
    def _boxes_intersect(
        start_a: List[float], shape_a: List[float], start_b: List[float], shape_b: List[float]
    ) -> bool:
        """Check if the two cubes intersect or not.

        Args:
            start_a (List[float]): bottom left corner of the cube A.
            shape_a (List[float]): size of each dimension in the cube A.
            start_b (List[float]): bottom left corner of the cube B.
            shape_b (List[float]): size of each dimension in the cube B.

        Returns:
            bool: True if the two cubes intersect. otherwise False.
        """

        if start_a[0] > start_b[0] + shape_b[0]:
            return False
        if start_b[0] > start_a[0] + shape_a[0]:
            return False
        if start_a[1] > start_b[1] + shape_b[1]:
            return False
        if start_b[1] > start_a[1] + shape_a[1]:
            return False
        if start_a[2] > start_b[2] + shape_b[2]:
            return False
        if start_b[2] > start_a[2] + shape_a[2]:
            return False
        return True

    @staticmethod
    def _get_union_areas(
        start_a: List[float], shape_a: List[float], start_b: List[float], shape_b: List[float]
    ) -> float:
        """Calculates the Union of the areas of the two cubes A and B.

        Args:
            start_a (List[float]): bottom left corner of the cube A.
            shape_a (List[float]): size of each dimension in the cube A.
            start_b (List[float]): bottom left corner of the cube B.
            shape_b (List[float]): size of each dimension in the cube B.

        Returns:
            float: union of the two cubes.
        """
        area_a = Metrics._get_area(shape_a)
        area_b = Metrics._get_area(shape_b)
        inter_area = Metrics._get_intersection_area(start_a, shape_a, start_b, shape_b)
        return float(area_a + area_b - inter_area)

    @staticmethod
    def _get_area(shape: List[float]) -> float:
        """calculates the area of a cube.

        Args:
            shape (List[float]): size of each dimension in the cube.

        Returns:
            float: area of the cube.
        """

        return math.prod(shape)

    @staticmethod
    def _get_intersection_area(
        start_a: List[float], shape_a: List[float], start_b: List[float], shape_b: List[float]
    ) -> float:
        """Calculates the intersection of the areas of the two cubes A and B.

        Args:
            start_a (List[float]): bottom left corner of the cube A.
            shape_a (List[float]): size of each dimension in the cube A.
            start_b (List[float]): bottom left corner of the cube B.
            shape_b (List[float]): size of each dimension in the cube B.

        Returns:
            float: intersection of the two cubes.
        """
        x_a = max(start_a[0], start_b[0])
        y_a = max(start_a[1], start_b[1])
        z_a = max(start_a[2], start_b[2])
        x_b = min(start_a[0] + shape_a[0], start_b[0] + shape_b[0])
        y_b = min(start_a[1] + shape_a[1], start_b[1] + shape_b[1])
        z_b = min(start_a[2] + shape_a[2], start_b[2] + shape_b[2])
        # intersection area
        return (x_b - x_a) * (y_b - y_a) * (z_b - z_a)

    @uval_stage
    def plot_roc_curves(self, roc_metrics: List[dict], show_graphic: bool = False) -> None:
        """Plot the ROC curve for every class.

        Args:
            roc_metrics (List[dict]): Output of some basic_metric. needs to be
            called before this method.
            show_graphic (bool, optional): if True, the plot will be shown. Defaults to False.

        Raises:
            IOError: [description]
        """

        result = None
        # Each resut represents a class
        for result in roc_metrics:
            if result is None:
                raise IOError("Error:No data for this class could be found.")

            class_id = result["Class"]

            recall = result["recall"]
            fpr = result["fpr"]

            plt.close()
            plt.plot(fpr, recall, label=f"ROC for IOU:{self.iou_threshold}")
            plt.xlabel("FP Rate")
            plt.ylabel("TP Rate")
            plt.title("ROC curve \nClass: %s" % str(class_id))
            plt.legend(shadow=True)
            plt.grid()
            plt.xlim([0.0, 1.0])
            plt.ylim([0.0, 1.0])
            if self.output_path is not None:
                plt.savefig(os.path.join(self.output_path, class_id + "_roc.png"))
            if show_graphic is True:
                plt.show(block=False)
                plt.pause(0.05)
        plt.close()



# -*- coding: utf-8 -*-
"""A StageData object acts like a data container that manages results generated by stages and read by subsequent
stages.

Depending on the stage, each class that inherits from StageData will contain different things, documented below.
Many of them will make use of pandas DataFrames to keep the data in a table-like shape.
"""

from abc import ABC
from typing import List

from pandas import DataFrame  # type: ignore

from uval.utils.log import logger


class StageData(ABC):
    """The abstract base class for all data that is passed around between stages.
    For now, we don't need anything here, but we will sure need this later,
    for example to store and restore the data when using caching."""

    pass


class TableStageData(StageData):
    """A subclass for all data that is table-like and can be handled using pandas"""

    frame_columns = []  # type: List[str]

    def __init__(self):
        self.table = DataFrame(columns=self.frame_columns)

    def to_html(self, file_path: str):
        self.table.to_html(file_path)

    def to_csv(self, file_path: str):
        self.table.to_csv(file_path)

    def from_dict_as_rows(self, data_dict: dict, check_data_validity: bool = False) -> bool:
        # Creates a DataFrame from a dictionary as a row for each dictionary key
        self.table = DataFrame.from_dict(data_dict, orient="index")

        # checking the number of rows after assignment
        if check_data_validity and len(data_dict) != self.table.shape[0]:
            logger.error("Unequal number of rows between the DataFrame and Dict after import!")
            # raise ValueError("Unequal number of rows")
            return False
        # checking the columns of the first row after assignment
        elif check_data_validity and len(data_dict.keys()) > 0:
            first_key = list(data_dict.keys())[0]
            if len(data_dict[first_key]) != self.table.shape[1]:
                logger.error("Unequal number of columns between the DataFrame and Dict after import!")
                # raise ValueError("Unequal number of columns")
                return False
        return True


class DatasetSpecificationData(TableStageData):
    """Contains a list of targets to be detected. Every entry / row refers to a single ground truth
    target. So multiple rows can refer to the same volume_id if there are multiple targets in one volume.

    for more information please see the The Data-split YAML format and HDF5 format in the uval docs.

    """

    frame_columns = ["volume_id", "label_id", "is_negative", "subset", "class_name"]

    def __init__(self):
        super().__init__()


class SupportedDatasetSpecificationData(DatasetSpecificationData):
    """Similar to DatasetSpecificationData, but in addition, every entry is supported
    by Hdf5 files for ground truth and detections"""

    frame_columns = DatasetSpecificationData.frame_columns + ["hdf5_detection", "hdf5_groundtruth", "target_id"]

    def __init__(self):
        super().__init__()


class Hdf5FilesData(TableStageData):
    """Keeps a list of files that are available on disk.
    The files can be identified by bag id.
    For each bag id, contains the file path and the file_meta and volume_meta information.
    Also stores detections and groundtruth and their scores and box data etc. but not the masks or projections."""

    frame_columns = ["volume_id", "file_path", "file_stat", "hdf5_meta"]

    def __init__(self):
        super().__init__()


class BagLevelAlarmsData(TableStageData):
    """For every bag id, stores the number of alarms on this bag"""

    frame_columns = ["volume_id", "num_alarms"]

    def __init__(self):
        super().__init__()


class DetectionMetricsData(TableStageData):
    """For every bag id, for every GT object, stores the matched detection alarm (index) and its overlap.
    In addition, for every bag, stores the detection alarm indices that do not have a matching ground truth"""

    pass



# -*- coding: utf-8 -*-
"""
    This module provides functions to verify uval HDF5 files.
    It reports any problems that a set of existing HDF5 may have.
    Any missing required field in HDF5 file will be reported as a problem
"""

from pathlib import Path
from typing import Dict, List

from uval.utils.hdf5_format import (
    DSET_VOLUME,
    h5_check_detection_fields,
    h5_check_file_meta_fields,
    h5_check_groundtruth_fields,
    h5_check_volcache,
    h5_check_volume_meta_fields,
)
from uval.utils.hdf5_io import UvalHdfFileInput


def verify_single_hdf5_file(file_path: str) -> list:
    """
        Checks a single hdf5 file and returns a list of error descriptions.
    Args:
        file_path: The HDF5 file path to be verified

    Returns:
        A list of problems detected in HDF5 file
    """
    problems = []

    # Check file name (extensions)
    lowercase_extensions = [s.lower() for s in Path(file_path).suffixes]
    extension_ok = False

    if lowercase_extensions[-1] == ".h5" and len(lowercase_extensions) >= 2:
        # Now check second level extension (second to last)
        if lowercase_extensions[-2] in [".det", ".gt", ".voldata", ".volcache"]:
            extension_ok = True

    if not extension_ok:
        return [
            "File must have a valid extension (one of '.det.h5', '.gt.h5', '.voldata.h5', '.volcache.h5'). "
            "Won't check its contents."
        ]

    with UvalHdfFileInput(file_path) as f:
        try:
            # Check file_meta
            h5_check_file_meta_fields(f.h5)
        except ValueError as e:
            problems += [str(e)]

        try:
            # Check volume_meta
            h5_check_volume_meta_fields(f.h5)
        except ValueError as e:
            problems += [str(e)]

        try:
            # Check contents depending on file name
            if lowercase_extensions[-2] == ".det":
                h5_check_detection_fields(f.h5)
            elif lowercase_extensions[-2] == ".gt":
                h5_check_groundtruth_fields(f.h5)
            elif lowercase_extensions[-2] == ".voldata":
                # Voldata is automatically checked in h5_check_volume_meta_fields above if available
                # Here we just check it's available at all
                if DSET_VOLUME not in f.h5:
                    raise ValueError("File does not contain volume data")
            elif lowercase_extensions[-2] == ".volcache":
                h5_check_volcache(f.h5)
        except ValueError as e:
            problems += [str(e)]

    return problems


def verify_hdf5_files(
    folder_path: str, recursive: bool = False, file_filter: str = "*.h5", print_problems: bool = True
) -> Dict[str, List[str]]:
    """
        Give a folder, finds and verifies all HDF5 files inside. This can be recursive or filtered if desired.
    Returns a dictionary with all the problems found for each file.
    Args:
        folder_path: The path to folder containing HDF5 files
        recursive: To parse the folder recursively or not
        file_filter: The wildcard to include HDF5 files by name
        print_problems: To print the detected problems in standard output or not

    Returns:
        A dictionary containing all the detected problems including the file name and problem description
    """

    # A dictionary mapping from file path to a list of strings with problem descriptions
    all_problems = {}

    # Find all HDF5 files that could be relevant
    if recursive:
        filepath_generator = Path(folder_path).glob(f"**/{file_filter}")
    else:
        filepath_generator = Path(folder_path).glob(file_filter)

    for file_path in filepath_generator:
        file_path_str = str(file_path.resolve())
        problems = verify_single_hdf5_file(file_path_str)
        if len(problems):
            all_problems[file_path_str] = problems

    if print_problems:
        for filepath in sorted(all_problems.keys()):
            print(f"\n{filepath}:")
            for p in all_problems[filepath]:
                print(f"- {p}")

    return all_problems



"""This module provides simple functions to read and write YAML files.
The data structure for pure YAML IO operations would be the dictionary
"""

from os import W_OK, access, path
from typing import Dict

from ruamel.yaml import YAML

from uval.utils.log import logger


def load_yaml_data(yaml_file_path: str) -> Dict:
    """
        loads data from a YAML file into a directory structure and returns dict type
        If the file does not exist or not read properly, returns None.
    Args:
        yaml_file_path: The path to the YAML file

    Returns:
        Data read from YAML file in dictionary format
    """
    if not path.exists(yaml_file_path):
        logger.critical("The YAML file '{}' does not exist!")
        return dict()
    try:
        yaml = YAML()
        with open(yaml_file_path, "r") as f:
            data_dict = yaml.load(f)
        return data_dict
    except Exception as e:
        logger.error("Could read YAML data from the file '{}': {}".format(yaml_file_path, e.__str__()))
        return dict()


def store_yaml_data(data_dict: dict, yaml_file_path: str) -> bool:
    """
        Stores a dictionary type data into YAML file and returns True
        If the folder is not writable of there's a problem with the writing, it returns False

    Args:
        data_dict:
        yaml_file_path:

    Returns:
        Returns true if YAML stored successfully, otherwise returns false
    """
    writing_dir = path.dirname(path.abspath(yaml_file_path))
    if not access(writing_dir, W_OK):
        logger.error("Unable to write into the destination folder '{}'!".format(writing_dir))
        return False

    yaml = YAML()
    try:
        with open(yaml_file_path, "w") as f:
            yaml.dump(data_dict, f)
        return True
    except Exception as e:
        logger.error("Could not write YAML data into the file '{}': {}".format(yaml_file_path, e.__str__()))
        return False



# -*- coding: utf-8 -*-

import os
import tempfile
import unittest

import h5py  # type: ignore
import numpy as np

from uval.utils.hdf5_format import DICTGROUP_GROUNDTRUTH, GROUP_FILE_META, GROUP_VOLUME_META, LISTGROUP_DETECTIONS
from uval.utils.hdf5_io import UvalHdfFileInput, UvalHdfFileOutput

SAMPLE_H5 = os.path.join(os.path.dirname(__file__), "sample.h5")

SAMPLE_DATA_DICT = {
    GROUP_FILE_META: {"det_version": "20-99-05"},
    GROUP_VOLUME_META: {
        "id": "BAG",
        "file_md5": "1a3ef6",
        "full_shape": np.array([1, 3, 2]),
        "is_cropped": 1,
        "roi_start": np.array([0, 0, 0]),
        "roi_shape": np.array([620, 420, 300]),
    },
    LISTGROUP_DETECTIONS: [
        {
            "class_name": "knife",
            "roi_start": np.array([0, 0, 0]),
            "roi_shape": np.array([10, 11, 12]),
            "score": 0.95,
            "mask": np.random.randint(2, size=(3, 3, 3)).astype(np.uint8),
        },
        {
            "class_name": "blunt",
            "roi_start": np.array([0, 0, 0]),
            "roi_shape": np.array([10, 11, 12]),
            "score": 0.98,
            "mask": np.random.randint(2, size=(3, 3, 3)).astype(np.uint8),
        },
    ],
    DICTGROUP_GROUNDTRUTH: {
        "%_label_1": {
            "class_name": "knife",
            "target_id": "",
            "roi_start": np.array([0, 0, 0]),
            "roi_shape": np.array([10, 11, 12]),
            "mask": np.random.randint(2, size=(3, 3, 3)).astype(np.uint8),
        },
        "%_label_2": {
            "class_name": "blunt",
            "target_id": "",
            "roi_start": np.array([0, 0, 0]),
            "roi_shape": np.array([10, 11, 12]),
            "mask": np.random.randint(2, size=(3, 3, 3)).astype(np.uint8),
        },
    },
}


def populate_test_hdf(out_h5: UvalHdfFileOutput) -> None:
    """
        Populates the object in param with sample data by reference
    Returns:
        None
    """
    out_h5.file_meta = SAMPLE_DATA_DICT[GROUP_FILE_META]
    out_h5.volume_meta = SAMPLE_DATA_DICT[GROUP_VOLUME_META]
    out_h5.detections = SAMPLE_DATA_DICT[LISTGROUP_DETECTIONS]
    out_h5.groundtruth = SAMPLE_DATA_DICT[DICTGROUP_GROUNDTRUTH]


def write_sample_h5_file(h5_path: str):
    """
        Writes an hdf5 data using the sample data
    Args:
        h5_path: Filepath to write h5 file into

    Returns:
        True if the file is created
    """
    with UvalHdfFileOutput(h5_path) as out_h5_f:
        populate_test_hdf(out_h5_f)
    if not h5py.is_hdf5(h5_path):
        raise ValueError("Sample file is not a valid hdf5 file")
    return os.path.isfile(h5_path)


class Hdf5TestSuite(unittest.TestCase):
    def setUp(self):
        self.tempdir_object = tempfile.TemporaryDirectory()
        self.tempdir = self.tempdir_object.name
        self.temp_h5 = os.path.join(self.tempdir, "test.h5")

    def test_writing_empty_hdf5(self):
        self.assertFalse(os.path.isfile(self.temp_h5), "File was there before creating it")
        # No volume metadata is set
        with self.assertRaises(ValueError):
            with UvalHdfFileOutput(self.temp_h5) as out_f:
                self.assertFalse(os.path.isfile(self.temp_h5), "Could not create HDF5 file")
                self.assertTrue(out_f.is_closed())
                out_f.write()
                self.assertFalse(out_f.is_closed())
        # File should not exist
        with self.assertRaises(OSError):
            os.unlink(self.temp_h5)

    def test_writing_hdf5(self):
        self.assertFalse(os.path.isfile(self.temp_h5), "File was there before creating it")

        with UvalHdfFileOutput(self.temp_h5) as out_f:
            self.assertFalse(os.path.isfile(self.temp_h5), "Could not create HDF5 file")
            populate_test_hdf(out_f)

        # Now verify the contents
        with h5py.File(self.temp_h5, "r") as f:
            self.assertTrue(GROUP_FILE_META in f)

            # label_group = f[HDFSchema.LABEL_GROUP]
            # num_labels = label_group[HDFSchema.NUM_LABELS][()]

        # Check if we can remove the file (if this fails, the file may still be open although it shouldn't)
        os.unlink(self.temp_h5)

    def test_reading_hdf5_all_at_once(self):
        self.assertTrue(write_sample_h5_file(SAMPLE_H5), f"Could not create sample h5 file: {SAMPLE_H5}")
        with UvalHdfFileInput(SAMPLE_H5) as f:
            print(f.read_all_fields())
        os.unlink(SAMPLE_H5)

    def test_reading_hdf5_step_by_step(self):
        # Check if file is closed in between
        self.assertTrue(write_sample_h5_file(SAMPLE_H5), f"Could not create sample h5 file: {SAMPLE_H5}")
        h5_sample = UvalHdfFileInput(SAMPLE_H5)
        print(h5_sample.volume_meta())
        self.assertTrue(h5_sample.is_closed())
        print(h5_sample.file_meta())
        self.assertTrue(h5_sample.is_closed())
        self.assertRaises(TypeError, h5_sample.ground_truth())
        self.assertTrue(h5_sample.is_closed())
        os.unlink(SAMPLE_H5)

    def test_integration_write_then_read1(self):
        with UvalHdfFileOutput(SAMPLE_H5) as out_h5_f:
            populate_test_hdf(out_h5_f)
            self.assertTrue(out_h5_f.is_closed())
        with UvalHdfFileInput(SAMPLE_H5) as in_h5_f:
            print(in_h5_f.volume_meta())
            self.assertFalse(in_h5_f.is_closed())
            print(in_h5_f.detections())
            self.assertFalse(in_h5_f.is_closed())
            print(in_h5_f.ground_truth())
            self.assertFalse(in_h5_f.is_closed())
        os.unlink(SAMPLE_H5)

    def test_integration_write_then_read2(self):
        # Test initializing from sample.h5 and writing back identically (as-is)
        # Then read again and compare all fields
        with UvalHdfFileOutput(SAMPLE_H5) as out_h5_f:
            populate_test_hdf(out_h5_f)
            self.assertTrue(out_h5_f.is_closed())
        with UvalHdfFileInput(SAMPLE_H5) as in_h5_f:
            self.assertEqual(in_h5_f.file_meta()["det_version"], SAMPLE_DATA_DICT[GROUP_FILE_META]["det_version"])
            sample_meta = SAMPLE_DATA_DICT[GROUP_VOLUME_META]
            for key in sample_meta.keys():
                if isinstance(sample_meta[key], (np.ndarray, np.generic)):
                    self.assertEqual(in_h5_f.volume_meta()[key].tolist(), sample_meta[key].tolist())
                else:
                    self.assertEqual(in_h5_f.volume_meta()[key], sample_meta[key])
        os.unlink(SAMPLE_H5)


if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-
"""This module provides functions to read and write uval HDF5 files.
For a format specification, have a look at `hdf5_format.py`
"""

import getpass  # type: ignore
import os  # type: ignore
import socket  # type: ignore
from abc import ABC  # type: ignore
from datetime import datetime, timezone  # type: ignore
from typing import Optional, Union  # type: ignore

import h5py  # type: ignore
from h5py import Group as H5Group  # type: ignore

from uval.utils.hdf5_format import *  # type: ignore
from uval.utils.log import logger  # type: ignore


class UvalHdfFile(ABC):
    def __init__(self, filepath: str, mode: str = "r"):
        self.filepath = filepath
        self.h5 = None
        self.file_mode = mode


class UvalHdfFileInput(UvalHdfFile):
    """A class to manage, read from a single uval-specific HDF5 file.
    The file on disk will not be held open by default.
    Every operation will open and close the file.
    If a bunch of operations is executed in a row and the file shall be held open, use
    a with-context on the UvalHdfFile object.
    """

    def __init__(self, filepath: str):
        super().__init__(filepath, "r")

        """Keeps track of nested with-statements.
           File will only be closed after exiting the last with-block"""
        self.context_counter = 0

    def __enter__(self):
        """
            Called on `with uval_file_obj:` to ensure the file is open.
            This is also used internally by the other member functions.
        Returns:
            The self object
        """
        self.context_counter += 1
        if self.h5 is None:
            try:
                self.h5 = h5py.File(self.filepath, self.file_mode)
            except OSError as err:
                logger.error(f"OS error '{err}' occurred while reading file: {self.filepath}")
                self.h5 = None
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """Ensure the h5 file on disk is closed when we leave the with-context"""
        self.context_counter -= 1

        if self.context_counter == 0 and self.h5 is not None:
            self.h5.close()
            self.h5 = None

    def is_closed(self):
        """Checks if the file is closed or not initialized"""
        return not self.h5.__bool__()

    def _join_path(self, *args) -> str:
        """
            joins multiple elements of an HDF5 group or dataset path to one string.
        Examples:
            ["root", "some/other", "field"] would map to "root/some/other/field"
        Args:
            *args: The elements to be joined

        Returns:
            The joined input elements as a string
        """

        return "/".join(args)

    def _convert_read(self, item: object):
        """
            Takes a data field read from an HDF5 file and makes sure it's in the right format,
            to pass it to python. Bytes for example will be decoded to unicode strings.
        Args:
            item: the object to be decoded after reading

        Returns:
            The decoded object to pass to python
        """

        if isinstance(item, bytes):
            return item.decode()
        else:
            return item

    def _read_listgroup_to_dict(
        self, listgroup: str, listitem_fields: Union[dict, List[str], str], start_path: str = ""
    ) -> Optional[list]:
        try:
            list_keys = self.h5[self._join_path(start_path, listgroup)].keys()  # type: ignore
        except (TypeError, KeyError):
            return None

        return [
            self._read_to_dict(listitem_fields, start_path=self._join_path(start_path, listgroup, list_key))
            for list_key in list_keys
        ]

    def _read_dictgroup_to_dict(
        self, dictgroup: str, dictitem_fields: Union[dict, List[str], str], start_path: str = ""
    ) -> Optional[dict]:
        try:
            dict_keys = self.h5[self._join_path(start_path, dictgroup)].keys()  # type: ignore
        except (TypeError, KeyError):
            return None

        return {
            dict_key: self._read_to_dict(dictitem_fields, start_path=self._join_path(start_path, dictgroup, dict_key))
            for dict_key in dict_keys
        }

    def _read_to_dict(self, fields: Union[dict, List[str], str], start_path: str = "") -> Optional[dict]:
        """
            Extract (copy) all the data for the given fields from the HDF5 file, and return them in a dictionary.
            This is done recursively.

        Args:
            fields: The fields in HDF5 file to extract data from
            start_path: The start path to start extracting requested data from
        Returns:
            A dictionary of data in the requested filed
        """
        result = {}
        with self:
            if start_path:
                try:
                    # Let's check if the start_path exists within the HDF5 file.
                    # If not, we return None
                    self.h5[start_path]  # type: ignore
                except (TypeError, KeyError):
                    return None

            if isinstance(fields, str):
                # Single string (reading one data set) will be transformed to a list with one element
                fields = [fields]

            if isinstance(fields, list):
                # To not replicate reading code, transform list into dict with None-values
                fields = {k: None for k in fields}

            if isinstance(fields, dict):
                for key, subfields in fields.items():
                    if subfields is None:
                        # In this case, the key is the name of the dataset to read
                        try:
                            result[key] = self._convert_read(
                                self.h5[self._join_path(start_path, key)][()]  # type: ignore
                            )  # type: ignore
                        except KeyError:
                            # Don't set if not available
                            # Don't throw an error either
                            pass
                    else:
                        cur_result = self._read_to_dict(subfields, start_path=self._join_path(start_path, key))
                        if cur_result is not None:
                            result[key] = cur_result

        return result

    def _read_to_dict_outer(self, fields: Union[dict, List[str], str], start_path: str = "") -> Optional[dict]:
        """
            A wrapper around _read_to_dict, which leaves away the first hierarchy level of the returned dict.
            That means, if you request fields like {'root': {'A': None, 'B': None}}, it would directly return the
            inner dict with A and B and it's values, but not the root node.

        Args:
            fields: The requested fields to be read from the HDF5 file
            start_path: The field to start reading from it

        Returns:
            The results read from the HDF5 file in a dictionary
        """
        if isinstance(fields, dict):
            assert len(fields.keys()) == 1

        result = self._read_to_dict(fields, start_path)
        if result:
            return result[next(iter(fields.keys()))]  # type: ignore

        return result

    def file_meta(self):
        return self._read_to_dict_outer(
            {
                GROUP_FILE_META: {
                    DSET_HOST_NAME: None,
                    DSET_USER_NAME: None,
                    DSET_DT_GENERATED: None,
                    DSET_DET_VERSION: None,
                }
            }
        )

    def volume_meta(self, include_caches=False):
        """
            The metadata information regarding the ct 3d image.
            Please refer to UVal hdf5 format.
        Args:
            include_caches: Whether to include the cached data or not

        Returns:
            None
        """
        if include_caches:
            return self._read_to_dict_outer(
                {
                    GROUP_VOLUME_META: {
                        DSET_ID: None,
                        DSET_FILE_MD5: None,
                        DSET_FULL_SHAPE: None,
                        DSET_IS_CROPPED: None,
                        DSET_ROI_SHAPE: None,
                        DSET_ROI_START: None,
                        GROUP_CACHE: {DSET_PROJECTION_X, DSET_PROJECTION_Y, DSET_PROJECTION_Z},
                    }
                }
            )
        else:
            return self._read_to_dict_outer(
                {
                    GROUP_VOLUME_META: {
                        DSET_ID: None,
                        DSET_FILE_MD5: None,
                        DSET_FULL_SHAPE: None,
                        DSET_IS_CROPPED: None,
                        DSET_ROI_SHAPE: None,
                        DSET_ROI_START: None,
                    }
                }
            )

    def volume(self):
        """
            The ct 3d volume stored in hdf5 file
            Please refer to UVal hdf5 format.
        Returns:
            None
        """
        return self._read_to_dict_outer(DSET_VOLUME)

    def ground_truth(self, include_masks=False, include_caches=False):
        """
            A list of 3d groundtruth data belonging to a ct image
            Please refer to UVal hdf5 format.
        Args:
            include_masks: To include 3d masks while reading or not
            include_caches: To include cached data while reading ot not

        Returns:
            None
        """
        item_descriptor = {
            DSET_CLASS_NAME: None,
            DSET_TARGET_ID: None,
            DSET_ROI_START: None,
            DSET_ROI_SHAPE: None,
        }
        if include_masks:
            item_descriptor.update({DSET_MASK: None})

        if include_caches:
            item_descriptor.update({GROUP_CACHE: {DSET_PROJECTION_X, DSET_PROJECTION_Y, DSET_PROJECTION_Z}})

        return self._read_dictgroup_to_dict(DICTGROUP_GROUNDTRUTH, item_descriptor)

    def detections(self, include_masks=False, include_caches=False):
        """
            A list of detected areas in the ct image
            Please refer to UVal hdf5 format.
        Args:
            include_masks: To include 3d masks while reading or not
            include_caches: To include cached data while reading ot not

        Returns:
            None
        """
        item_descriptor = {
            DSET_CLASS_NAME: None,
            DSET_ROI_START: None,
            DSET_ROI_SHAPE: None,
            DSET_SCORE: None,
        }
        if include_masks:
            item_descriptor.update({DSET_MASK: None})

        if include_caches:
            item_descriptor.update({GROUP_CACHE: {DSET_PROJECTION_X, DSET_PROJECTION_Y, DSET_PROJECTION_Z}})

        return self._read_listgroup_to_dict(LISTGROUP_DETECTIONS, item_descriptor)

    def read_all_fields(self) -> dict:
        """
            Reads all the existing fields in h5 file
        Returns:
            All the groups in a dictionary
        """
        return {
            GROUP_FILE_META: self.file_meta(),
            GROUP_VOLUME_META: self.volume_meta(),
            DSET_VOLUME: self.volume(),
            DICTGROUP_GROUNDTRUTH: self.ground_truth(),
            LISTGROUP_DETECTIONS: self.detections(),
        }


class UvalHdfFileOutput(UvalHdfFile):
    """A class to manage, write to a single uval-specific HDF5 file.
    The file on disk will not be held open by default.
    Every operation will open and close the file.
    If a bunch of operations is executed in a row and the file shall be held open, use
    a with-context on the UvalHdfFile object.
    """

    _volume: Optional[np.ndarray]
    _volume_meta: Optional[dict]
    _groundtruth: Optional[dict]
    _detections: Optional[list]
    _file_meta: Optional[dict]

    # Always keeps all things in memory before writing
    def __init__(self, filepath: str, copy_from_input: Optional[UvalHdfFileInput] = None):
        super().__init__(filepath, "w")
        if copy_from_input:
            self.read_all_from(copy_from_input)
        else:
            self._volume = None
            self._volume_meta = None
            self._groundtruth = None
            self._detections = None
            self._file_meta = None

    def __enter__(self):
        """
            For the output class, we only actually write when done, not at this point.
        Returns:
            Self object of type UvalHdfFileOutput
        """
        if not os.access(os.path.abspath(os.path.dirname(self.filepath)), os.W_OK):
            raise IOError(f"Cannot write to file '{self.filepath}'")
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """We are done, we can write the file now, if no exception occurred."""
        if exc_type is None:
            # Do not write if an exception occurred
            self.write()

    def _on_close(self):
        """Before closing, write the changes"""
        self.write()

    def _write_dict_to_group(self, h5group: H5Group, fields: dict):
        """In the HDF5 file, create groups and field at given start_path, containing the data from fields.

        Args:
            h5group: The HDF5 group to write the filed into
            fields: A nested dictionary representing the data to be written
        Returns:
            None
        """

        # The file needs to be writable
        assert self.file_mode == "w"

        for k, v in fields.items():
            if isinstance(v, dict):
                h5group.create_group(k)
                self._write_dict_to_group(h5group[k], v)
            else:
                h5group.create_dataset(k, data=v)

    def write(self):
        """
            Explicitly call this method to write the file. However, changes will automatically
            be written when the file is closed.
        Returns:
            None
        """

        # Check for meta data
        if self._volume_meta is None:
            raise ValueError(
                "No volume metadata was set before writing. " "Uval files always must contain volume metadata."
            )

        with h5py.File(self.filepath, "w") as f:
            # Write file meta data (This will always be generated here, cannot be set manually)
            group_file_meta = f.create_group(GROUP_FILE_META)
            group_file_meta.create_dataset(DSET_HOST_NAME, data=socket.gethostname())
            group_file_meta.create_dataset(DSET_USER_NAME, data=getpass.getuser())
            group_file_meta.create_dataset(
                DSET_DT_GENERATED, data=str(datetime.now(timezone.utc).astimezone().isoformat())
            )

            # As stated above, file_meta cannot be set manually - with one exception: det_version
            try:
                group_file_meta.create_dataset(DSET_DET_VERSION, data=self._file_meta[DSET_DET_VERSION])
            except (TypeError, KeyError):
                group_file_meta.create_dataset(DSET_DET_VERSION, data="not specified")

            # Write volume meta data
            check_volume_meta_fields(self._volume_meta, self._volume)
            group_volume_meta = f.create_group(GROUP_VOLUME_META)
            self._write_dict_to_group(group_volume_meta, self._volume_meta)

            # Write volume data if available
            if self._volume is not None:
                f.create_dataset(DSET_VOLUME, data=self._volume)

            # Write detections
            if self._detections:
                check_detection_fields(self._detections)
                listgroup_detections = f.create_group(LISTGROUP_DETECTIONS)
                for idx, detection in enumerate(self._detections):
                    group_detection = listgroup_detections.create_group(str(idx))
                    self._write_dict_to_group(group_detection, detection)

            # Write groundtruth:
            if self._groundtruth:
                check_groundtruth_fields(self._groundtruth)
                listgroup_groundtruth = f.create_group(DICTGROUP_GROUNDTRUTH)
                for label_name, gt in self._groundtruth.items():
                    group_gt = listgroup_groundtruth.create_group(str(label_name))
                    self._write_dict_to_group(group_gt, gt)

    def is_closed(self):
        """Checks if the file is closed or not initialized"""
        return not self.h5.__bool__()

    def read_all_from(self, input_file: UvalHdfFileInput) -> None:
        """Reading all the meta data included in another HDF5 file
        Args:
            input_file: HDF file to read from

        Returns:
            None
        """

        self._groundtruth = input_file.ground_truth()
        self._detections = input_file.detections()
        self._file_meta = input_file.file_meta()
        self._volume_meta = input_file.volume_meta()
        self._volume = input_file.volume()

    # File_meta
    @property
    def file_meta(self):
        raise IOError("This class should only be used to write, not to read a uval file")

    @file_meta.setter
    def file_meta(self, value: dict):
        self._file_meta = value

    @file_meta.deleter
    def file_meta(self):
        self._file_meta = None

    # Volume
    @property
    def volume(self):
        raise IOError("This class should only be used to write, not to read a uval file")

    @volume.setter
    def volume(self, value: np.ndarray):
        self._volume = value

    @volume.deleter
    def volume(self):
        self._volume = None

    # Volume meta
    @property
    def volume_meta(self):
        raise IOError("This class should only be used to write, not to read a uval file")

    @volume_meta.setter
    def volume_meta(self, value: dict):
        self._volume_meta = value

    @volume_meta.deleter
    def volume_meta(self):
        self._volume_meta = None

    # Detections
    @property
    def detections(self):
        raise IOError("This class should only be used to write, not to read a uval file")

    @detections.setter
    def detections(self, value: list):
        self._detections = value

    @detections.deleter
    def detections(self):
        self._detections = None

    # Groundtruth
    @property
    def groundtruth(self):
        raise IOError("This class should only be used to write, not to read a uval file")

    @groundtruth.setter
    def groundtruth(self, value: list):
        self._groundtruth = value  # type: ignore

    @groundtruth.deleter
    def groundtruth(self):
        self._groundtruth = None



import os
from typing import Dict, List, cast

import jinja2
import pandas as pd  # type: ignore
from matplotlib import pyplot as plt  # type: ignore

from uval.stages.stage import uval_stage  # type: ignore
from uval.utils.log import logger


class DiffReporter:
    """This class implements a comparison for atleast two methods.
    The basic UVal analysis should have been performed beforehand.
    """

    def __init__(self, models: List[Dict]) -> None:

        self.models = models
        if all(["iou_range" in model.keys() for model in models]):
            self.report_type = "RANGE"
        else:
            self.report_type = "SINGLE"
        name = "_".join([str(result.get("title")) for result in models])

        self.output_path = os.path.join(".", name)
        os.makedirs(os.path.join(".", name), exist_ok=True)
        threshold = {model.get("iou_threshold") for model in models}
        assert len(threshold) == 1
        self.iou_threshold = threshold.pop()
        self.classes = sorted([result["Class"] for result in models[0]["single_threshold"]])
        self.template_file = "template_range_diff.html" if self.report_type == "RANGE" else "template_diff.html"
        self.templates_path = "src/uval/templates"
        self.report_file = "report_diff.html"

    def run(self):
        """This function performs the comparison. plots and generates the report."""
        self.plot_roc_curves(self.models)
        self.plot_precision_recall_curve(self.models)
        if self.report_type == "RANGE":
            self.plot_recall_iou_curve(self.models)
        self.generate_report(self.models)

    @uval_stage
    def plot_roc_curves(self, models: List[Dict]) -> None:
        """Plot the ROC curve for every class.

        Args:
            models (List[dict]): Output of Uval analysis for at least two methods.
        """

        result = None
        models_single: List[list] = [cast(list, model.get("single_threshold")) for model in models]
        classes = {result["Class"] for model in models_single for result in model}
        n_cls = {len(model) for model in models_single}
        # All models should have the same classes in the results
        assert len(n_cls) == 1
        n_classes = n_cls.pop()
        assert len(classes) == n_classes
        # Each result represents a class
        for class_id in classes:
            plt.close()
            for model_single, model in zip(models_single, models):
                contestant = model.get("title")
                result = next(filter(lambda x: x["Class"] == class_id, model_single))
                recall = result["recall"]
                fpr = result["fpr"]
                plt.plot(fpr, recall, label=f"ROC for IOU:{self.iou_threshold}, model:{contestant}")

            plt.xlabel("FP Rate")
            plt.ylabel("TP Rate")
            plt.title("ROC curve \nClass: %s" % str(class_id))
            plt.legend(shadow=True)
            plt.grid()
            plt.xlim([0.0, 1.0])
            plt.ylim([0.0, 1.0])
            if self.output_path is not None:
                plt.savefig(os.path.join(self.output_path, class_id + "_roc.png"))
            plt.close()

    @uval_stage
    def plot_precision_recall_curve(self, pascal_voc_metrics: List[Dict]) -> None:
        """Plot the Precision x Recall curve for a given class.

        Args:
            models (List[dict]): Output of Uval analysis for at least two methods.
        """

        def value_picker(x, idx, field):
            return x[idx].get(field)

        # Each result represents a class
        for cls in self.classes:
            plt.close()
            for model in pascal_voc_metrics:
                contestant = model["title"]
                idx = [c["Class"] for c in model["single_threshold"]].index(cls)
                precision = value_picker(model["single_threshold"], idx, "precision")
                recall = value_picker(model["single_threshold"], idx, "recall")
                average_precision = value_picker(model["single_threshold"], idx, "AP")
                plt.plot(
                    recall,
                    precision,
                    label=(
                        f"Precision for IOU:{self.iou_threshold},"
                        + f"Model:{contestant},"
                        + f"AP={(average_precision):{5}.{3}}"
                    ),
                )

            plt.xlabel("recall")
            plt.ylabel("precision")

            plt.title("Precision x Recall curve \nClass:" + cls)
            plt.legend(shadow=True)
            plt.grid()

            plt.xlim([0.0, 1.0])
            plt.ylim([0.0, 1.0])
            plt.savefig(os.path.join(self.output_path, cls + "_precision_recall.png"))

            plt.close()

    @uval_stage
    def plot_recall_iou_curve(self, models: List[Dict]) -> None:
        """Plot the Recall x IOU curve for a given class.

        Args:
            models (List[dict]): Output of Uval analysis for at least two methods.
        """
        # Each result represents a class
        for cls in self.classes:

            plt.close()
            for model in models:
                contestant = model["title"]
                idx = [c["Class"] for c in model["single_threshold"]].index(cls)

                iou_thresholds = model["iou_range"]
                recall_vector = [model.get("rs").get(iou)[idx] for iou in iou_thresholds]  # type:ignore
                ap_str = "{0:.2f}%".format(sum(recall_vector) / len(recall_vector) * 100)

                plt.plot(
                    iou_thresholds,
                    recall_vector,
                    label=f"Confidence:{model.get('confidence_threshold')}, Model: {contestant}, AP: {ap_str}",
                )
            plt.xlabel("IOU")
            plt.ylabel("Recall")

            plt.title(f"Recall x IOU curve \nClass: {cls}")

            plt.legend(shadow=True)
            plt.grid()
            plt.xlim([0.0, 1.0])
            plt.ylim([0.0, 1.0])
            plt.savefig(os.path.join(self.output_path, cls + "_recall_iou.png"))
            plt.close()

    @uval_stage
    def generate_report(self, models: List[Dict]) -> None:
        """_summary_

        Args:
            models (List[Dict]): Result of UVal analysis on atleast two methods.
        """
        # Sample DataFrame
        range_results = [dict(model) for model in models]
        single_results = [result.pop("single_threshold") for result in range_results]
        contestants = [str(result.get("title")) for result in range_results]

        def func(row):

            highlight = "background-color: darkorange;"
            default = ""

            return [default] * (len(row) - 1) + [highlight]

        classes = self.classes
        to_remove = ["precision", "recall", "fpr", "interpolated precision", "interpolated recall"]
        [r.pop(field) for field in to_remove for res in single_results for r in res]
        # [r.pop("Class") for r in res]
        # for key, value in kwargs.items():
        high_level = dict()
        to_remove = ["ap", "ars", "iou_range"]
        [r.pop(field, None) for field in to_remove for r in range_results]
        rs = [r.pop("rs", None) for r in range_results]
        high_level["mar"] = [r.pop("mar", None) for r in range_results]

        cell_hover = {  # for row hover use <tr> instead of <td>
            "selector": "td:hover",
            "props": [("background-color", "#ffffb3")],
        }
        row_hover = {  # for row hover use <tr> instead of <td>
            "selector": "tr:hover",
            "props": [("background-color", "#ffffb3")],
        }

        sorted_idx = []
        single_results_sorted = []
        for model in single_results:
            m_cls = [m.get("Class") for m in model]
            s_idx = [i[0] for i in sorted(enumerate(m_cls), key=lambda x: x[1])]
            single_results_sorted.append([model[i] for i in s_idx])
            sorted_idx.append(s_idx)

        pd_index = [val for val in contestants for _ in range(len(classes))]
        columns = list(single_results_sorted[0][0].keys())

        df = pd.DataFrame(
            [r.values() for result in single_results_sorted for r in result], index=pd.Index(pd_index), columns=columns
        )

        styler = (
            df.style.set_caption(f"Calculated metrics for iou:{self.iou_threshold}")
            # .set_precision(2)
            .format(precision=2).set_table_styles([row_hover])
        )
        env = jinja2.Environment(loader=jinja2.FileSystemLoader(searchpath=self.templates_path))
        template = env.get_template(self.template_file)
        total_images = []

        for cls in classes:
            img_names = []
            img_names.append("." + "/" + cls + "_roc.png")
            img_names.append("." + "/" + cls + "_precision_recall.png")
            if self.report_type == "RANGE":
                img_names.append("." + "/" + cls + "_recall_iou.png")
            total_images.append(img_names)

        # TODO if exists
        if self.report_type == "RANGE":
            for result in range_results:
                result["map"]["Total"] = sum(result["map"].values()) / len(result["map"])
            columns = list(range_results[0]["map"].keys())
            df2 = pd.DataFrame(
                [result["map"].values() for result in range_results],
                index=pd.Index(contestants),
                columns=[str(round(c, 2)) if isinstance(c, float) else c for c in columns],
            )
            rs_sorted = []
            for k in range(len(sorted_idx)):
                rs_sorted.append({key: [val[i] for i in sorted_idx[k]] for key, val in rs[k].items()})

            pd_index = [val for val in contestants for _ in range(len(classes))]
            pd_lines = []
            for c in range(len(classes)):
                for result in rs_sorted:
                    pd_lines.append([classes[c]] + [v[c] for v in result.values()])
            df_rs = pd.DataFrame(
                pd_lines, index=pd.Index(pd_index), columns=["Class"] + [str(e) for e in list(rs_sorted[0].keys())]
            )

            styler_rs = (
                df_rs.style.set_caption("Recall values for all classes and all IOU thresholds")
                .format(precision=2)
                .set_table_styles([row_hover])
            )

            styler2 = (
                df2.style.set_caption("Mean average precision for various IOU levels.")
                .format(precision=2)
                .set_table_styles([cell_hover])
                # .apply(func, subset=["Total"], axis=1)
            )

            # Template handling

            html = template.render(
                range_table=styler2.to_html(),
                single_table=styler.to_html(),
                rs_table=styler_rs.to_html(),
                total_images=total_images,
                # mar=round(high_level["mar"][0], 2),
                mar=" vs. ".join(
                    [contestants[r] + ":" + str(round(high_level["mar"][r], 2)) for r in range(len(range_results))]
                ),
                map=" vs. ".join(
                    [
                        contestants[r] + ":" + str(round(range_results[r]["map"]["Total"], 2))
                        for r in range(len(range_results))
                    ]
                ),
                # map=round(high_level["map"][0], 2),
                title=" x ".join(contestants),
            )
        else:
            html = template.render(
                single_table=styler.to_html(), total_images=total_images, title=" x ".join(contestants)
            )
            # Template handling

        # Write the HTML file
        with open(os.path.join(self.output_path, self.report_file), "w") as f:
            f.write(html)
        logger.info(f"Report saved to {os.path.join(self.output_path, self.report_file)}.")



# -*- coding: utf-8 -*-


class Hdf5Virtual:
    """
    This module provides the class Hdf5Virtual which represents a reference to an Hdf5 file
    that actually resides on disk somewhere. It keeps meta data and other low-storage information in memory.
    When trying to access larger parts of the file, like volume, masks or projections, it accesses the actual file.
    """

    def __init__(self):
        msg = f"Method '{print.__name__}' not implemented in class '{self.__class__}'."
        raise NotImplementedError(msg)



# -*- coding: utf-8 -*-
"""This module provides the Context class. The user of uval shall create a single instance
of this class and call all the stages as needed.

For an example on how to use this, please that a look at the examples folder.
"""

from contextlib import AbstractContextManager

_uval_instance = None


class Context:
    def __init__(self, max_workers=None):
        self.config = {"hdf5_io": {"num_threads": max_workers}}
        self._cached = False  # Set to true when in a cached context (with ctx.cached(): block)
        self.cache_folder = None
        self.cached_context_manager = CachedContextManager(self)

    def set_cache_folder(self, folder_path: str):
        self.cache_folder = folder_path

    def cached(self):
        if self.cache_folder is None:
            raise ValueError("Before using cached, you need to set a folder with `Context.set_cache_folder()`")
        return self.cached_context_manager


class CachedContextManager(AbstractContextManager):
    """This context manager will be returned by `Context.cached()` and can be used in a with
    statement. All"""

    _stream = None

    def __init__(self, ctx: Context):
        self.ctx = ctx
        self.entry_counter = 0

    def __enter__(self):
        self.entry_counter += 1
        self.ctx._cached = True

    def __exit__(self, exctype, excinst, exctb):
        self.entry_counter -= 1
        if self.entry_counter <= 0:
            self.entry_counter = 0
            self.ctx._cached = False


def get_context(max_workers):
    global _uval_instance
    if _uval_instance is None:
        _uval_instance = Context(max_workers)
    return _uval_instance



from uval.stages.combine_files import create_supported_dataset_from_intersection, support_dataset_with_file_paths
from uval.stages.dataset_specification import load_datasplit
from uval.stages.hdf5 import load_evaulation_files
from uval.stages.metrics import Metrics
from uval.stages.stage import uval_stage  # type: ignore



# -*- coding: utf-8 -*-
"""This utils module currently only provides logging functionality.
Once this grows to much, we will need to split it.
"""

import logging

from rich.logging import RichHandler

FORMAT = "%(message)s"
logging.basicConfig(level=logging.INFO, format=FORMAT, datefmt="[%X]", handlers=[RichHandler(rich_tracebacks=True)])
logger = logging.getLogger("uval")


class RootLogger:
    """
    All the child logger messages will propagate through this root logger
    """

    def __init__(self, logging_level=logging.DEBUG):
        """
            The root logger will be configured but should never be used directly for logging
        Child loggers will propagate their messages up to the root logger
        Args:
            logging_level: The verbosity level of logger
        """

        self.root_logger = logging.getLogger()
        self.root_logger.setLevel(logging_level)

        # The logging formats are currently not configurable
        self.stream_formatter = logging.Formatter(fmt="%(asctime)s %(name)s [%(levelname)s]: %(message)s")
        self.popup_formatter = logging.Formatter(fmt="%(name)s [%(levelname)s]: %(message)s")

        # Now set up our initial logging handlers
        self.set_up_handlers()

    def set_up_handlers(self):
        """
        Sets up Stream Handlers
        """
        stream_handler = logging.StreamHandler()
        stream_handler.setFormatter(self.stream_formatter)
        self.root_logger.handlers = [
            stream_handler,
        ]



# type: ignore
"""This module defines the abstract base class for a stage.
A stage is one step of an evaluation pipeline to be executed, for example
converting detections from one format to another.
"""

import functools
from typing import Callable

from uval.utils.log import logger


def uval_stage(func: Callable):
    """This decorator has to be applied to every stage that is defined.
    We can use this to log the start and end of stages, handle caching etc."""

    @functools.wraps(func)
    def wraps_stage(*args, **kwargs):
        logger.info(f"Starting stage '{func.__qualname__}'")
        result = func(*args, **kwargs)
        logger.info(f"Stage '{func.__qualname__}' done")
        return result

    return wraps_stage



import unittest
from os import path

import pytest

from uval.config import default_argument_parser, get_cfg
from uval.stages.combine_files import support_dataset_with_file_paths
from uval.stages.dataset_specification import load_datasplit
from uval.stages.hdf5 import load_evaulation_files
from uval.stages.metrics import Metrics

DATA_PATH = "./data"
datasplit_subset = ["train", "test"]


@pytest.fixture
def setup():
    """
    Create configs and perform basic setups.
    """
    parser = default_argument_parser()
    args = parser.parse_args()
    cfg = get_cfg()
    if args.config_file:
        cfg.merge_from_file(args.config_file)
    cfg.merge_from_list(args.opts)
    cfg.freeze()
    return cfg


@pytest.fixture
def load_test_data(setup):
    print(setup)
    dataset = load_datasplit(path.join(DATA_PATH, "datasplit/uval_ds.yaml"), datasplit_subset)
    hdf5_detections, hdf5_groundtruth, soc_data = load_evaulation_files(path.join(DATA_PATH, "hdf5"), dataset=dataset)
    supported_dataset = support_dataset_with_file_paths(hdf5_groundtruth, hdf5_detections, soc_data)
    return Metrics(supported_dataset, setup.METRICS, setup.OUTPUT)


def test_basic_metric(load_test_data):
    print("iou th is:", load_test_data.iou_threshold)
    print("conf th is:", load_test_data.confidence_threshold)
    print("is there data?:", load_test_data.data)
    basic_metric = load_test_data.basic_metric()
    print(basic_metric)
    assert basic_metric is not None
    all_classes = set([metric["Class"] for metric in basic_metric])
    assert len(basic_metric) == len(all_classes)  # are classes unique?
    for metric in basic_metric:
        assert (min(metric["precision"]) >= 0) & (max(metric["precision"]) <= 1)
        assert (min(metric["recall"]) >= 0) & (max(metric["recall"]) <= 1)
        assert len(metric["precision"]) == len(metric["recall"])
        assert metric["Total Positives"] > 0
        assert metric["Total Negatives"] > 0
        assert metric["Total TP"] <= metric["Total Positives"]
        assert metric["Total FP"] <= metric["Total Negatives"]
        assert metric["Total FN"] <= metric["Total Positives"]
        assert metric["Total TP soft"] <= metric["Total TP"]
        assert metric["Total FP soft"] <= metric["Total FP"]
        assert metric["Total FN soft"] >= metric["Total FN"]


def test_get_pascal_voc2012_metric(load_test_data):
    voc_007 = load_test_data.get_pascal_voc2007_metric()
    voc_012 = load_test_data.get_pascal_voc2012_metric()
    for metric7, metric12 in zip(voc_007, voc_012):
        assert metric7["AP"] > 0 and metric7["AP"] < 1
        assert metric12["AP"] > 0 and metric12["AP"] < 1
        assert metric7["AP"] == pytest.approx(metric12["AP"], 0.25)


def test_get_fscore(load_test_data):
    basic_metrics11 = load_test_data.basic_metric(iou_threshold=0.1, confidence_threshold=0.1)
    fscore_metric11 = load_test_data.get_fscore(basic_metrics11)
    for metric in fscore_metric11:
        assert metric["F score"] >= 0 and metric["F score"] <= 1
        assert metric["F score soft"] >= 0 and metric["F score soft"] <= 1

    basic_metrics91 = load_test_data.basic_metric(iou_threshold=0.9, confidence_threshold=0.1)
    for metric11, metric91 in zip(basic_metrics11, basic_metrics91):
        assert metric11["Total TP"] >= metric91["Total TP"]

    basic_metrics15 = load_test_data.basic_metric(iou_threshold=0.1, confidence_threshold=0.9)
    for metric11, metric15 in zip(basic_metrics11, basic_metrics15):
        assert metric11["Total TP"] > metric15["Total TP"]


def test_generate_report(load_test_data):
    load_test_data.evaluate()
    assert path.isfile(path.join(load_test_data.output_path, load_test_data.report_file))


class TestIOU(unittest.TestCase):
    def test_area(self):
        assert Metrics._get_area([1, 2, 3]) == 6

    def test_intersection_area(self):
        assert Metrics._get_intersection_area([0, 0, 0], [2, 3, 4], [1, 1, 1], [3, 2, 1]) == 2

    def test_union_area(self):
        assert Metrics._get_union_areas([0, 0, 0], [2, 3, 4], [1, 1, 1], [3, 2, 1]) == 28

    def test_intersect(self):
        assert Metrics._boxes_intersect([0, 0, 0], [1, 1, 1], [1, 1, 1], [1, 1, 1]) is True
        assert Metrics._boxes_intersect([0, 0, 0], [1, 1, 1], [2, 1, 1], [1, 1, 1]) is False
        assert Metrics._boxes_intersect([0, 0, 0], [1, 1, 1], [1, 2, 1], [1, 1, 1]) is False
        assert Metrics._boxes_intersect([0, 0, 0], [1, 1, 1], [1, 1, 2], [1, 1, 1]) is False
        assert Metrics._boxes_intersect([3, 0, 0], [1, 1, 1], [1, 1, 1], [1, 1, 1]) is False
        assert Metrics._boxes_intersect([0, 3, 0], [1, 1, 1], [1, 1, 1], [1, 1, 1]) is False
        assert Metrics._boxes_intersect([0, 0, 3], [1, 1, 1], [1, 1, 1], [1, 1, 1]) is False

    def test_iou(self):
        assert Metrics.iou([-1, 0, 0], [2, 3, 4], [1, 1, 1], [3, 2, 1]) == 0
        assert Metrics.iou([3, 0, 0], [1, 1, 1], [1, 1, 1], [1, 1, 1]) == 0
        assert Metrics.iou([1, 1, 1], [2, 3, 4], [2, 2, 2], [3, 2, 1]) == (2.0 / (30.0 - 2.0))



import pickle
from os import path

from uval.config import default_argument_parser
from uval.config.config_utils import setup_from_args
from uval.main import single
from uval.stages.diff_reporter import DiffReporter

CONFIGS_PATH_1 = "./output/example/config.yaml,./output/example2/config.yaml"
CONFIGS_PATH_2 = "./output/example2/config.yaml,./output/example2/config.yaml"


"""
def test_generate_report(load_test_data):
    load_test_data.evaluate()
    assert path.isfile(path.join(load_test_data.output_path, load_test_data.report_file))
"""


def test_diffs():
    diffs(CONFIGS_PATH_1)
    diffs(CONFIGS_PATH_2)


def diffs(configs):
    parser = default_argument_parser()
    args = parser.parse_args(["--config-file", configs])
    # args = parser.parse_args()
    print("Command Line Args:", args)

    cfgs = setup_from_args(args)
    assert len(cfgs) == 2
    for cfg in cfgs:
        single(cfg)

    if len(cfgs) > 1:
        print("initiating the generation of a differential report...")
        # the contents of the dataset and the splits should match
        # otherwise comparison is meaningless
        # the metrics should also match

        results = []
        for cfg in cfgs:
            filename = path.join(cfg.OUTPUT.PATH, cfg.OUTPUT.METRICS_FILE)
            with open(filename, "rb") as f:
                results.append(pickle.load(f))
        reporter = DiffReporter(results)
        reporter.run()
        assert path.isdir(reporter.output_path)
        assert path.isfile(path.join(reporter.output_path, reporter.report_file))



"""This module provides functions to read and write uval HDF5 files.
The supported HDF5 fields defined listed as yaml:

File name suggestion (let's keep tools functional even if names not complying)
- NAME.det.h5  (does not contain volume_data, groundtruth)
- NAME.gt.h5  (does not contain detections, volume_data)
- NAME.volcache.h5 (only volume_meta including volume projection cache)
- NAME.voldata.h5 (only volume_data)

# X, Y and Z axis
# Z-axis is belt direction (dir. of motion)
# Y-axis is vertical pointing up
# X-axis is point left when looking in belt motion direction

# Character set always UTF-8

file_meta:  # Always required
  host_name: "H5T_STRING"  # Host name of computer that generated the h5 file e.g. philscomputer
  user_name: "H5T_STRING"  # User name of user that generated the h5 file
  dt_generated: "H5T_STRING"  # ISO 8601 time and date of file creation (with timezone!)

volume_meta:  # Always available, not optional! (also for dets and gt)
  id: "H5T_STRING"  # e.g. BAGGAGE_20181122_081331_126018
  file_md5: "H5T_STRING"  # The checksum of the original ct volume file e.g. 686593fa1f05f610066129b72c62bfdd
  full_shape: INT (3) # Shape of full volume
  is_cropped: INT  # if 1, the data only contains the voxels within the roi
  roi_start: INT (3)
  roi_shape: INT (3) # Matches size of data if "is_cropped" is True
  cache:  # Optional
    projection_x: UINT8 RGB IMAGE  # Colored Matlum if possible, otherwise grayscale (R=G=B)
    projection_y: UINT8 RGB IMAGE
    projection_z: UINT8 RGB IMAGE

volume_data: "H5T_STD_U16LE"  # Optional (to save space, not contained in .dets.h5 and .gt.h5)

detections:  # list int-indexed as strings for each member (e.g. "0", "1", ..)
  class_name: "H5T_STRING"
  roi_start: INT (3)
  roi_shape: INT (3) # Same as size of mask if mask is available
  mask: "H5T_STD_U8LE"  # Optional (e.g. only bounding boxes)
  score: FLOAT
  cache:  # Optional
    density: FLOAT
    mass: FLOAT
    num_voxels: INT
    projection_x: UINT16 # Taking 3D mask with 1s and 0s, adding up along x axis (only y and z axis remain)
    projection_y: UINT16 # Taking 3D mask with 1s and 0s, adding up along y axis (only x and z axis remain)
    projection_z: UINT16 # Taking 3D mask with 1s and 0s, adding up along z axis (only x and y axis remain)

groundtruth:  # dict indexed by label id for each member
  class_name: "H5T_STRING"
  target_id: "H5T_STRING" # Formerly known as threat id
  roi_start: INT (3)
  roi_shape: INT (3) # Same as size of mask if mask is available
  mask: "H5T_STD_U8LE" # Optional (e.g. only bounding boxes)
  cache:  # Optional
    projection_x:  # As for detections
    projection_y:
    projection_z:


# 2D MASK projection to 1D
# X ---->
# 0 0 0 0 0 0 0  ^
# 0 1 0 1 0 0 0  |
# 0 1 1 1 1 0 0  |
# 0 0 1 1 0 0 0  Y
# 0 0 0 1 0 0 0

# 0 2 2 4 1 0 0  Projection (adding up)
# 0 1 1 1 1 0 0  Binary mask

# Proj along Y
"""

from enum import Enum
from typing import Any, Dict, List, Union

import numpy as np
from h5py import Dataset as H5Dataset  # type: ignore
from h5py import File as H5File  # type: ignore
from h5py import Group as H5Group  # type: ignore
from h5py import check_string_dtype

# Having all HDF5 internal group and field names as constants, helps preventing typos
# because a typo will make python complain on execution

GROUP_FILE_META = "file_meta"
GROUP_VOLUME_META = "volume_meta"
GROUP_CACHE = "cache"
LISTGROUP_DETECTIONS = "detections"
DICTGROUP_GROUNDTRUTH = "groundtruth"

DSET_HOST_NAME = "host_name"
DSET_USER_NAME = "user_name"
DSET_DT_GENERATED = "dt_generated"
DSET_DET_VERSION = "det_version"

DSET_ID = "id"
DSET_FILE_MD5 = "file_md5"

DSET_VOLUME = "volume"
DSET_FULL_SHAPE = "full_shape"
DSET_IS_CROPPED = "is_cropped"
DSET_ROI_START = "roi_start"
DSET_ROI_SHAPE = "roi_shape"
DSET_CLASS_NAME = "class_name"
DSET_SUBCLASS_NAME = "subclass_name"
DSET_TARGET_ID = "target_id"
DSET_MASK = "mask"
DSET_SCORE = "score"

DSET_PROJECTION_X = "projection_x"
DSET_PROJECTION_Y = "projection_y"
DSET_PROJECTION_Z = "projection_z"

DSET_DENSITY = "density"
DSET_MASS = "mass"
DSET_NUM_VOXELS = "num_voxels"


class FieldRequired(Enum):
    Required = 1
    Optional = 2


class ArrayReqs:
    """Used to represent requirements on an np.ndarray.
    Very similar to a type hint. So that `ArrayReqs(shape=(3,3,-1))` corresponds to a type hint like
    `np.ndarray[shape=(3,3,-1)]` where -1 indicates any size along that dimension.
    You can also use ArrayReqs(shape=3) to indicate the number of dimensions should be 3."""

    def __init__(self, shape=None, dtype=None):
        self.shape = shape
        self.dtype = np.dtype(dtype) if dtype else None

    def check(self, array: Union[np.ndarray, H5Dataset]) -> bool:
        if self.dtype and (array.dtype != self.dtype):
            return False

        if self.shape is None:
            return True

        if isinstance(self.shape, int):
            # Shape just is the number of dimensions
            # In this case, that's all we need to check and then return
            return self.shape == len(array.shape)
        else:
            # Shape is actually the shape (tuple with length for each dimension)
            if len(self.shape) != len(array.shape):
                return False

        # We now check every dimension for the length specified. -1 means any length is fine.
        for dim_idx in range(len(self.shape)):
            if self.shape[dim_idx] >= 0 and self.shape[dim_idx] != array.shape[dim_idx]:
                return False
        # If we passed all checks, the array is good
        return True

    def __repr__(self):
        if self.dtype is not None:
            return f"<ArrayReqs shape={self.shape} dtype={self.dtype}>"
        else:
            return f"<ArrayReqs shape={self.shape}>"


# The following definition of requirements is used to:
# 1. Verify parts of a file when reading it
# 2. Verify parts to be written to a file before writing them

format_requirements = {
    GROUP_FILE_META: (
        {
            DSET_HOST_NAME: (str, FieldRequired.Required),
            DSET_USER_NAME: (str, FieldRequired.Required),
            DSET_DT_GENERATED: (str, FieldRequired.Required),
            DSET_DET_VERSION: (str, FieldRequired.Required),
        },
        FieldRequired.Required,
    ),
    GROUP_VOLUME_META: (
        {
            DSET_ID: (str, FieldRequired.Required),
            DSET_FILE_MD5: (str, FieldRequired.Required),
            DSET_FULL_SHAPE: (ArrayReqs((3,)), FieldRequired.Required),
            DSET_IS_CROPPED: (int, FieldRequired.Required),
            DSET_ROI_START: (ArrayReqs((3,)), FieldRequired.Required),
            DSET_ROI_SHAPE: (ArrayReqs((3,)), FieldRequired.Required),
            GROUP_CACHE: (
                {
                    DSET_PROJECTION_X: (np.ndarray, FieldRequired.Required),
                    DSET_PROJECTION_Y: (np.ndarray, FieldRequired.Required),
                    DSET_PROJECTION_Z: (np.ndarray, FieldRequired.Required),
                },
                FieldRequired.Optional,
            ),
        },
        FieldRequired.Required,
    ),
    DSET_VOLUME: (ArrayReqs(3, dtype=np.uint16), FieldRequired.Optional),
    LISTGROUP_DETECTIONS: (
        {
            DSET_CLASS_NAME: (str, FieldRequired.Required),
            DSET_SUBCLASS_NAME: (str, FieldRequired.Optional),
            DSET_ROI_START: (ArrayReqs((3,)), FieldRequired.Required),
            DSET_ROI_SHAPE: (ArrayReqs((3,)), FieldRequired.Required),
            DSET_MASK: (np.ndarray, FieldRequired.Optional),
            DSET_SCORE: (float, FieldRequired.Required),
            GROUP_CACHE: (
                {
                    DSET_DENSITY: (float, FieldRequired.Optional),
                    DSET_MASS: (float, FieldRequired.Optional),
                    DSET_NUM_VOXELS: (int, FieldRequired.Optional),
                    DSET_PROJECTION_X: (np.ndarray, FieldRequired.Required),
                    DSET_PROJECTION_Y: (np.ndarray, FieldRequired.Required),
                    DSET_PROJECTION_Z: (np.ndarray, FieldRequired.Required),
                },
                FieldRequired.Optional,
            ),
        },
        FieldRequired.Optional,
    ),
    DICTGROUP_GROUNDTRUTH: (
        {
            DSET_CLASS_NAME: (str, FieldRequired.Required),
            DSET_SUBCLASS_NAME: (str, FieldRequired.Optional),
            DSET_TARGET_ID: (str, FieldRequired.Required),
            DSET_ROI_START: (ArrayReqs((3,)), FieldRequired.Required),
            DSET_ROI_SHAPE: (ArrayReqs((3,)), FieldRequired.Required),
            DSET_MASK: (np.ndarray, FieldRequired.Optional),
            GROUP_CACHE: (
                {
                    DSET_PROJECTION_X: (np.ndarray, FieldRequired.Required),
                    DSET_PROJECTION_Y: (np.ndarray, FieldRequired.Required),
                    DSET_PROJECTION_Z: (np.ndarray, FieldRequired.Required),
                },
                FieldRequired.Optional,
            ),
        },
        FieldRequired.Optional,
    ),
}


def check_dataset_type(dataset: object, type_descriptor) -> bool:
    """Check for a single dataset (a python value or H5Dataset) if it matches the type requirement."""
    # Find type kind of given data
    # (https://numpy.org/doc/stable/reference/generated/numpy.dtype.kind.html#numpy.dtype.kind)
    if isinstance(dataset, H5Dataset):
        type_kind = dataset.dtype.kind
        if type_kind == "O" and check_string_dtype(dataset.dtype):
            type_kind = "S"  # It's a string, although technically not like numpy type 'S'
            # In the future we might also want to check whether its utf-8 and variable length
            # For that see https://docs.h5py.org/en/stable/special.html#variable-length-strings
    elif isinstance(dataset, str):
        type_kind = "S"
    elif isinstance(dataset, bool):
        type_kind = "b"
    elif isinstance(dataset, int):
        type_kind = "i"
    elif isinstance(dataset, float):
        type_kind = "f"
    elif isinstance(dataset, np.ndarray):
        type_kind = "nd"
    else:
        type_kind = "<None>"

    dataset_is_array = hasattr(dataset, "shape") and len(dataset.shape) > 0
    # Now check type descriptor against that type
    if isinstance(type_descriptor, ArrayReqs):
        if not dataset_is_array:
            # This is not an array
            return False
        return type_descriptor.check(dataset)
    elif type_descriptor is str:
        return type_kind in "SU" and not dataset_is_array
    elif type_descriptor is int:
        return type_kind in "iu" and not dataset_is_array
    elif type_descriptor is bool:
        return type_kind == "b" and not dataset_is_array
    elif type_descriptor is float:
        return type_kind == "f" and not dataset_is_array
    elif type_descriptor is np.ndarray:
        return type_kind == "nd" or dataset_is_array
    else:
        raise NotImplementedError(f"Type checking for {type_descriptor} not supported. Please implement it.")


def check_fields(to_check: Union[dict, H5Group], requirements: Union[Dict[Any, Any], Any], base_name: str = ""):
    """Checks a dict (group) or value (dataset) against requirements. Fields may be required or optional.
    Additional unknown fields will also result in a failed check.

    The requirements dict has to be defined as follows:
    Every entry in the dict maps from the field name to a 2-tuple (data_type, is_required),
    where `data_type` can be a python type like str, int, np.ndarray or ArrayWithShape to specify the
    data shape.
    The element `is_required` is an enum (see FieldRequired) which specifies if this field is optional.

    The requirements can be nested to represent groups. For this purpose, use a requirements
    dictionary as the `data_type` of one of the fields.

    Args:
        to_check: Nested dictionary to check against requirements
        requirements: As explained above
        base_name: String describing the location of `to_check` within the file (for better error messages only)
    """
    if isinstance(to_check, H5Group):
        available_datasets = {k for k in to_check if isinstance(to_check[k], H5Dataset)}
        available_groups = {k for k in to_check if isinstance(to_check[k], H5Group)}
    else:
        available_datasets = {k for k in to_check if not isinstance(to_check[k], dict)}
        available_groups = {k for k in to_check if isinstance(to_check[k], dict)}

    for field, requirement in requirements.items():
        is_group = isinstance(requirement[0], dict)

        if is_group:
            if field in available_groups:
                # Found a matching group (may or may not be optional)
                # Check recursively
                check_fields(to_check[field], requirement[0], base_name=f"{base_name}/{field}")
                available_groups.remove(field)
            else:
                if requirement[1] is FieldRequired.Required:
                    raise ValueError(f"Missing required group '{field}' in '{base_name}'")
        else:
            # This is a dataset not a group
            if field in available_datasets:
                if not check_dataset_type(to_check[field], requirement[0]):
                    raise ValueError(
                        f"Dataset '{field}' in '{base_name}' does not match required type '{requirement[0]}'"
                    )
                available_datasets.remove(field)
            else:
                if requirement[1] is FieldRequired.Required:
                    raise ValueError(f"Missing required dataset '{field}' in '{base_name}'")

    # Check if any available fields remain
    if len(available_groups) > 0:
        raise ValueError(f"Unknown groups {available_groups} in '{base_name}'")

    if len(available_datasets) > 0:
        raise ValueError(f"Unknown datasets {available_datasets} in '{base_name}'")


def check_listgroup_fields(listgroup: List[dict], requirements: Union[dict, Any], base_name: str = "") -> None:
    """Checks a list of instances against the requirements.
    The requirements apply to each item of the list, not to the list as a whole.
    Inside the HDF5 file, lists will be represented as groups containing multiple groups that
    have integers as names (but stored as string, because names have to be strings).
    In the native python dict format, we will use simple lists.
    """

    for item in listgroup:
        check_fields(item, requirements=requirements, base_name=base_name)


def h5_check_listgroup_fields(listgroup: H5Group, requirements: Union[dict, object], base_name: str = "") -> None:
    """Checks an H5Group that makes a list of instances against the requirements.
    The requirements apply to each item of the list, not to the list as a whole.
    Inside the HDF5 file, lists will be represented as groups containing multiple groups that
    have integers as names (but stored as string, because names have to be strings).
    """

    item_count = len(listgroup.keys())
    for idx in range(item_count):
        # We check if the group contains exactly the items named like "0", "1", ... "N"
        if str(idx) not in listgroup:
            raise ValueError(f"List group at {base_name} has {item_count} items but is missing index '{str(idx)}'")

        # Now check each item in the list if it's compliant with requirements
        check_fields(listgroup[str(idx)], requirements=requirements, base_name=base_name)


def check_dictgroup_fields(dictgroup: Dict[str, dict], requirements: Union[dict, object], base_name: str = "") -> None:
    """Checks a list-like dict of instances against the requirements.
    Almost like `check_listgroup_fields` but the elements are indexed by keys.
    The requirements apply to each item of the dict, not to the dict as a whole.
    """

    for item in dictgroup.values():
        check_fields(item, requirements=requirements, base_name=base_name)


def h5_check_dictgroup_fields(dictgroup: H5Group, requirements: Union[dict, object], base_name: str = "") -> None:
    """Checks an H5Group of similar sub-items against the requirements.
    Almost like `check_listgroup_fields` but the elements are indexed by keys.
    The requirements apply to each item of the group, not to the group as a whole.
    """

    for item in dictgroup.values():
        check_fields(item, requirements=requirements, base_name=base_name)


def check_file_meta_fields(file_meta: dict) -> None:
    check_fields(
        file_meta, requirements=format_requirements[GROUP_FILE_META][0], base_name=GROUP_FILE_META
    )  # type: ignore


def check_detection_fields(detections: list) -> None:
    check_listgroup_fields(
        detections,
        requirements=format_requirements[LISTGROUP_DETECTIONS][0],
        base_name=LISTGROUP_DETECTIONS,  # type: ignore
    )


def check_groundtruth_fields(groundtruth: dict) -> None:
    check_dictgroup_fields(
        groundtruth,
        requirements=format_requirements[DICTGROUP_GROUNDTRUTH][0],
        base_name=DICTGROUP_GROUNDTRUTH,  # type: ignore
    )


def check_volume_meta_fields(volume_meta: dict, volume: np.ndarray = None) -> None:
    check_fields(
        volume_meta, requirements=format_requirements[GROUP_VOLUME_META][0], base_name=GROUP_VOLUME_META
    )  # type: ignore

    # Check if meta data matches the volume, if we have one
    if volume is not None:
        if volume.shape != tuple(volume_meta[DSET_FULL_SHAPE]):
            raise ValueError("Volume shape does not match full shape given in meta data")


def h5_check_file_meta_fields(h5: H5File) -> None:
    try:
        file_meta_group = h5[GROUP_FILE_META]
    except KeyError:
        raise ValueError("Missing file_meta group")
    check_fields(
        file_meta_group, requirements=format_requirements[GROUP_FILE_META][0], base_name=GROUP_FILE_META
    )  # type: ignore


def h5_check_detection_fields(h5: H5File) -> None:
    try:
        detections_group = h5[LISTGROUP_DETECTIONS]
    except KeyError:
        raise ValueError("Missing detections group")
    h5_check_listgroup_fields(
        detections_group,
        requirements=format_requirements[LISTGROUP_DETECTIONS][0],
        base_name=LISTGROUP_DETECTIONS,  # type: ignore
    )


def h5_check_groundtruth_fields(h5: H5File) -> None:
    try:
        groundtruth_group = h5[DICTGROUP_GROUNDTRUTH]
    except KeyError:
        raise ValueError("Missing groundtruth group")
    h5_check_dictgroup_fields(
        groundtruth_group,
        requirements=format_requirements[DICTGROUP_GROUNDTRUTH][0],
        base_name=DICTGROUP_GROUNDTRUTH,  # type: ignore
    )


def h5_check_volume_meta_fields(h5: H5File) -> None:
    try:
        volume_meta_group = h5[GROUP_VOLUME_META]
    except KeyError:
        raise ValueError("Missing volume_meta group")

    check_fields(
        volume_meta_group, requirements=format_requirements[GROUP_VOLUME_META][0], base_name=GROUP_VOLUME_META
    )  # type: ignore

    # Check if meta data matches the volume, if we have one
    if DSET_VOLUME in h5:
        if not check_dataset_type(h5[DSET_VOLUME], format_requirements[DSET_VOLUME][0]):
            raise ValueError(
                f"Volume does not match required type '{format_requirements[DSET_VOLUME][0]}'. "
                f"Its type is {h5[DSET_VOLUME]}"
            )

        if tuple(h5[DSET_VOLUME].shape) != tuple(volume_meta_group[DSET_FULL_SHAPE]):
            raise ValueError("Volume shape does not match full shape given in meta data")


def h5_check_volcache(h5: H5File) -> None:
    """Checks the volume cache, which is usually optional. Here we require it"""

    try:
        volcache_group = h5[GROUP_VOLUME_META][GROUP_CACHE]
    except KeyError:
        raise ValueError("Missing volume cache group")

    check_fields(
        volcache_group,
        requirements=format_requirements[GROUP_VOLUME_META][0][GROUP_CACHE][0],  # type: ignore
        base_name=f"{GROUP_VOLUME_META}/{GROUP_CACHE}",
    )



# -*- coding: utf-8 -*-
"""This module include stages that are used to find and load HDF5 files from disk.
The result will be returned as a Hdf5FilesData object
"""

import warnings  # type: ignore
from concurrent.futures import ThreadPoolExecutor, as_completed  # type: ignore
from glob import glob  # type: ignore
from os import path, stat  # type: ignore
from typing import Tuple  # type: ignore

from rich.progress import Progress

from uval.context import get_context  # type: ignore
from uval.stages.stage import uval_stage  # type: ignore
from uval.stages.stage_data import DatasetSpecificationData, Hdf5FilesData  # type: ignore
from uval.utils.hdf5_io import UvalHdfFileInput  # type: ignore
from uval.utils.log import logger  # type: ignore


def _read_single_file_meta(file_path: str):
    with UvalHdfFileInput(file_path) as f:
        return file_path, {
            "file_meta": f.file_meta(),
            "volume_meta": f.volume_meta(),
            "detections": f.detections(include_masks=False, include_caches=False),
            "groundtruth": f.ground_truth(include_masks=False, include_caches=False),
        }


def load_hdf5_files(
    folder_path: str,
    recursive: bool = False,
    file_filter: str = "*.h5",
    list_of_files=None,
    ignore_missing_files=False,
    max_workers=None,
) -> Hdf5FilesData:
    """This function searches for uval compatible HDF5 files in the given folder.
    It will load their meta data to make sure we can match and access them quickly by volume id."""

    config = get_context(max_workers=max_workers).config["hdf5_io"]

    # Set up result data store
    stage_results = Hdf5FilesData()

    # Find all HDF5 files that could be relevant
    if list_of_files is None or recursive:
        if recursive:
            file_list = glob(path.join(folder_path, "**", file_filter), recursive=True)
        else:
            file_list = glob(path.join(folder_path, file_filter), recursive=False)
    else:
        file_list = list_of_files
    logger.debug("file_list")

    all_futures = set()
    with ThreadPoolExecutor(max_workers=config["num_threads"]) as executor:

        # Submit all jobs to the pool
        for file_name in file_list:
            all_futures.add(executor.submit(_read_single_file_meta, file_name))
        """
        for file_name in file_list:
            if path.exists(file_name):
                all_futures.add(executor.submit(_read_single_file_meta, file_name))
            elif ignore_missing_files:
                logger.warning(f"HDF5 file does not exist: '{file_name}'")
            else:
                raise IOError(f"HDF5 file does not exist: '{file_name}'")
        """
        # Wait for each to complete
        for future in as_completed(all_futures):
            file_path, meta = future.result()

            # Store meta data into results StageData in a way that it is easily accessible by its bag_id
            try:
                volume_id = meta["volume_meta"]["id"]
                stage_results.table.loc[volume_id] = {
                    "volume_id": volume_id,
                    "file_path": file_path,
                    "file_stat": stat(file_path),  # We need this to check if the file changed
                    "hdf5_meta": meta,
                }
            except KeyError as e:

                # We found an invalid HDF5 file, log this
                logger.warning(f"HDF5 file is missing some fields: '{file_path}'")
                raise (e)

    return stage_results


@uval_stage
def load_gt(
    folder_path: str, recursive: bool = False, dataset=None, ignore_missing_files=False, max_workers=None
) -> Hdf5FilesData:
    """Ground truth in non negative images are loaded based on the YAML file.

    Args:
        folder_path (str): path to the data.
        recursive (bool, optional): set to True if subdirectories should also be searched. Defaults to False.
        dataset ([DatasetSpecificationData], optional): The dataset that was loaded from YAML files. Defaults to None.
        ignore_missing_files: If a HDF5 file is missing, no error is raised

    Raises:
        Exception: if a file includes in the dataset but not in the directory (sanity check).

    Returns:
        Hdf5FilesData: Ground truths in positive images.
    """
    if dataset is None:
        warnings.warn("Old usage!")
    list_of_files = []
    if dataset is not None:
        list_of_files = dataset.table[~dataset.table.is_negative]["volume_id"].tolist()

    file_list_full = [path.join(folder_path, file + ".gt.h5") for file in list_of_files]
    try:
        gt_files = load_hdf5_files(
            folder_path,
            recursive,
            file_filter="*.gt.h5",
            list_of_files=file_list_full,
            ignore_missing_files=ignore_missing_files,
            max_workers=max_workers,
        )
    except FileNotFoundError as e:
        raise Exception("Ground Truth files and YAML mismatch.") from e
    return gt_files


@uval_stage
def load_detections(
    folder_path: str, recursive: bool = False, dataset=None, ignore_missing_files=False, max_workers=None
) -> Hdf5FilesData:
    """Detections in non negative images are loaded based on the YAML file.

    Args:
        folder_path (str): path to the data.
        recursive (bool, optional): set to True if subdirectories should also be searched. Defaults to False.
        dataset ([DatasetSpecificationData], optional): The dataset that was loaded from YAML files. Defaults to None.
        ignore_missing_files: If a HDF5 file is missing, no error is raised

    Raises:
        Exception: if a file includes in the dataset but not in the directory (sanity check).

    Returns:
        Hdf5FilesData: Detections in positive images.
    """
    if dataset is None:
        warnings.warn("Old usage!")
    list_of_files = []

    if dataset is not None:
        list_of_files = dataset.table[~dataset.table.is_negative]["volume_id"].tolist()
    file_list_full = [path.join(folder_path, file + ".det.h5") for file in list_of_files]
    try:
        files_det = load_hdf5_files(
            folder_path,
            recursive,
            file_filter="*.det.h5",
            list_of_files=file_list_full,
            ignore_missing_files=ignore_missing_files,
            max_workers=max_workers,
        )
    except FileNotFoundError as e:
        raise Exception("Detection files and YAML mismatch.") from e
    return files_det


@uval_stage
def load_negatives(
    folder_path: str,
    recursive: bool = False,
    dataset: DatasetSpecificationData = None,
    ignore_missing_files: bool = False,
    max_workers=None,
) -> Hdf5FilesData:
    """Detections in negative images are loaded based on the YAML file.

    Args:
        folder_path (str): path to the data.
        recursive (bool, optional): set to True if subdirectories should also be searched. Defaults to False.
        dataset ([DatasetSpecificationData], optional): The dataset that was loaded from YAML files. Defaults to None.
        ignore_missing_files: If a HDF5 file is missing, no error is raised

    Raises:
        Exception: if a file includes in the dataset but not in the directory (sanity check).

    Returns:
        Hdf5FilesData: Detections in Negative images.
    """
    list_of_files = []
    if dataset is not None:
        list_of_files = dataset.table[dataset.table.is_negative]["volume_id"].tolist()
    file_list_full = [path.join(folder_path, file + ".det.h5") for file in list_of_files]

    try:
        files_soc = load_hdf5_files(
            folder_path,
            recursive,
            file_filter="*.det.h5",
            list_of_files=file_list_full,
            ignore_missing_files=ignore_missing_files,
            max_workers=max_workers,
        )
    except FileNotFoundError as e:
        raise Exception("Negative detection files and YAML mismatch.") from e
    return files_soc


@uval_stage
def load_evaulation_files(
    folder_path: str, recursive=False, dataset=None, ignore_missing_files=False, max_workers=None
) -> Tuple[Hdf5FilesData, Hdf5FilesData, Hdf5FilesData]:
    """This is a wrapper that calls underlying functions and loads positive
    detections, GTs and negative detections.

    Args:
        folder_path (str): path to the data folder.
        recursive (bool, optional): set to True if subdirectories should also be searched. Defaults to False.
        dataset ([DatasetSpecificationData], optional): The input dataset
        loaded from YAML file. Defaults to None.
        ignore_missing_files: If a HDF5 file is missing, no error is raised

    Returns:
        Hdf5FilesData: Tuple including all three parts of the dataset.
    """
    neg = set(dataset.table[dataset.table.is_negative]["volume_id"].tolist())
    pos = set(dataset.table[~dataset.table.is_negative]["volume_id"].tolist())
    if len(neg & pos) > 0:
        raise ValueError("The positive and negative subsets have overlap.")

    with Progress() as progress:
        task = progress.add_task("[red]Loading...", total=3)
        detections: Hdf5FilesData = load_detections(
            path.join(folder_path, "detections"), recursive, dataset, ignore_missing_files, max_workers=max_workers
        )
        progress.update(task, advance=1)
        groundtruths: Hdf5FilesData = load_gt(
            path.join(folder_path, "raw"), recursive, dataset, ignore_missing_files, max_workers=max_workers
        )
        progress.update(task, advance=1)
        negatives: Hdf5FilesData = load_negatives(
            path.join(folder_path, "detections"), recursive, dataset, ignore_missing_files, max_workers=max_workers
        )
        progress.update(task, advance=1)
    return detections, groundtruths, negatives



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['uval', 'uval.config', 'uval.stages', 'uval.tests', 'uval.utils']

package_data = \
{'': ['*'], 'uval': ['templates/*']}

install_requires = \
['Jinja2>=3.0.3,<4.0.0',
 'Pillow>=9.0.1,<10.0.0',
 'click>=8.0.4,<9.0.0',
 'fvcore>=0.1.5,<0.2.0',
 'h5py>=3.6.0,<4.0.0',
 'ipykernel>=6.9.2,<7.0.0',
 'matplotlib>=3.5.1,<4.0.0',
 'numpy>=1.22.3,<2.0.0',
 'pandas>=1.4.1,<2.0.0',
 'pipx>=1.0.0,<2.0.0',
 'randomname>=0.1.5,<0.2.0',
 'rich>=12.0.1,<13.0.0',
 'ruamel.yaml>=0.17.21,<0.18.0',
 'scikit-learn>=1.0.2,<2.0.0',
 'sphinx-rtd-theme>=1.0.0,<2.0.0',
 'tqdm>=4.63.0,<5.0.0']

entry_points = \
{'console_scripts': ['uval = uval.main:main']}

setup_kwargs = {
    'name': 'uval',
    'version': '0.1.7',
    'description': 'This python package is meant to provide a high level interface to facilitate the evaluation of object detection and segmentation algorithms that operate on 3D volumetric data.',
    'long_description': '&nbsp;![UVal](https://gitlab.com/smithsdetection/uval/-/raw/main/icon_uval.png) &nbsp;&nbsp;&nbsp;&nbsp;&nbsp;[![Smiths Detection](https://gitlab.com/smithsdetection/uval/-/raw/main/SD_logo.png)](https://www.smithsdetection.com/ "Redirect to homepage")\n---\nUVal - Unified eValuation framework for 3D X-ray data\n---\n> This python package is meant to provide a high level interface to facilitate the evaluation of object detection and segmentation algorithms that operate on 3D volumetric data.\n---\n- There is a growing need for high performance detection algorithms using 3D data, and it is very important to be able to compare them. By far, there has not been a trivial solution for a straightforward comparison between different 3D detection algorithms.\n- This framework seeks a way to address the aforementioned problem by introducing a simple and standard layout of the popular HDF5 data format as input. \n- Each detection algorithm can export the results and groundtruth data according to the defined layout principles. Then, UVal can evaluate the performance and come up with common comparison metrics.\n\n| ![](3d_vol.gif "3d CT Volume") |  ![](https://gitlab.com/smithsdetection/uval/-/raw/main/dets_anim.gif "Detections") |\n| :---: | :---: |\n\n## Installation (non-development)\nIf you are not developing and only using UVal, you can simply install it as a `pypi` package; simply run:\n```shell\npip install uval\n```\n\nIf you would like to have UVal installation to be independant of a specific python environment, simply use `pipx` instead of `pip`.\n\nTo run the code you can type:\n```shell\nuval --config-file ${workspaceFolder}/output/example/config.yaml\n```\nFor an example of the outputs see [here](https://gitlab.com/smithsdetection/uval/-/tree/main/output/example) and the report [here](https://gitlab.com/smithsdetection/uval/-/raw/main/output/example/report.pdf).\n\nFor the details of each entry in the config file please see [here](https://gitlab.com/smithsdetection/uval/-/raw/main/src/uval/config/README.md).\n\n## Installation (development)\nWe recommend using the Anaconda package manager to create <em>virtual environment</em> for UVal:\n```shell\n# for Linux installation:\nwget https://repo.anaconda.com/archive/Anaconda3-2020.11-Linux-x86_64.sh\nsh ./Anaconda3-2020.11-Linux-x86_64.sh\n```\n\n## Development setup\n\n* First, please clone the UVal\'s git repository by executing the following command:  \n  ```git clone https://gitlab.com/smithsdetection/uval.git```\n  \n\n* You will need a `python >= 3.8` environment to develop Uval.  \n  We recommend using Anaconda due to its ease of use:\n  ```shell\n  wget https://repo.anaconda.com/archive/Anaconda3-2020.11-Linux-x86_64.sh\n  sh ./Anaconda3-2020.11-Linux-x86_64.sh\n  ```\n* Setting up a `conda virtual environment` with `python=3.8` using the following commands:\n  Install conda-lock with pip support.\n  ```shell\n  cd uval\n  conda-lock install --name UVALENV conda-lock.yml\n  poetry install\n  pre-commit install\n  ```\n  ALternatively, you can create your own conda environment from scratch. and follow with `poetry` and `pre-commit` installations.\n\n## Example code\n* A step-by-step walkthrough for reading and evaluating data with the UVal is available as a jupyter document:\n  * [jupyter notebook demo](https://gitlab.com/smithsdetection/uval/-/blob/main/demo/sample-data-evaluation.ipynb)\n------\n  * **Hint:** Prior to running the demo jupyter notebook walkthrough, the following steps must be performed:\n    \n    * The `ipykernel` conda package must be installed\n      ```shell\n      conda install -c anaconda ipykernel\n      ```\n    * The `uvalenv` environment must be added as an ipykernel: \n      ```shell  \n      python3 -m ipykernel install --user --name uvalenv --display-name "uvalenv Python38"\n      ```\n    * The `uvalenv Python38` kernel, which includes all the required python packages must be selected in `jupyter` environment to run the code.\n------\n\n## Documentations\nRead the docs: https://uval.readthedocs.io/\n\n## Release History\n\n* 0.1.x\n  * The first ready-to-use version of UVal releases publicly\n\n## Meta\n\nSmiths Detection – [@Twitter](https://twitter.com/smithsdetection) – uval@smithsdetection.com\n\n``UVal`` is released under the [GPL V3.0 license](LICENSE).\n\n## Contributing\n\n1. Fork it (<https://gitlab.com/smithsdetection/uval/fork>)\n2. Create your feature branch (`git checkout -b feature/fooBar`)\n3. Commit your changes (`git commit -am \'Add some fooBar\'`)\n4. Push to the branch (`git push origin feature/fooBar`)\n5. Create a new merge Request\n\n## Citing UVal\nIf you use UVal in your research or wish to refer to the results, please use the following BibTeX entry.\n\n```****BibTeX****\n@misc{smithsdetection2022uval,\n  author =       {Philipp Fischer, Geert Heilmann, Mohammad Razavi, Faraz Saeedan},\n  title =        {UVal},\n  howpublished = {\\url{https://gitlab.com/smithsdetection/uval}},\n  year =         {2022}\n}\n```',
    'author': 'Geert Heilmann',
    'author_email': 'geert.heilmann@smithsdetection.com',
    'maintainer': 'Faraz Saeedan',
    'maintainer_email': 'faraz.saeedan@smithsdetection.com',
    'url': 'https://gitlab.com/smithsdetection/uval',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



import hashlib
import pickle
from os import path

from rich.console import Console
from rich.traceback import install

from uval import *
from uval.stages.diff_reporter import DiffReporter
from uval.utils.log import logger
from uval.utils.yaml_io import load_yaml_data

console = Console()
install()


def main():
    parser = default_argument_parser()
    args = parser.parse_args()
    print("Command Line Args:", args)

    cfgs = setup_from_args(args)
    for cfg in cfgs:
        single(cfg)

    if len(cfgs) > 1:
        print("initiating the generation of a differential report...")
        # the contents of the dataset and the splits should match
        # otherwise comparison is meaningless
        # the metrics should also match
        md5_hash = {
            hashlib.md5(
                bytes(
                    str(dict(cfg.METRICS).pop("MAX_PROCESSES"))
                    + "".join(cfg.DATA_SPLIT.SUBSET)
                    + str(load_yaml_data(cfg.DATA_SPLIT.YAML)),
                    "UTF-8",
                )
            ).digest()
            for cfg in cfgs
        }

        if not len(set(md5_hash)) == 1:
            logger.warning("UVAL proudly refuses to compare Apples and Oranges!")
        else:
            logger.debug("The two datasets appear to be perfect matches")
            results = []
            for cfg in cfgs:
                filename = path.join(cfg.OUTPUT.PATH, cfg.OUTPUT.METRICS_FILE)
                with open(filename, "rb") as f:
                    results.append(pickle.load(f))
        print("loaded")
        reporter = DiffReporter(results)
        reporter.run()


def single(cfg):
    if path.isfile(path.join(cfg.OUTPUT.PATH, cfg.OUTPUT.METRICS_FILE)):
        print("metrics pickle already exists. skipping the evaluation...")
        return
    print(cfg)
    ctx = get_context(cfg.DATA.MAX_THREADS)
    ctx.set_cache_folder(cfg.ENV.CACHE_FOLDER)

    with ctx.cached():
        dataset = load_datasplit(cfg.DATA_SPLIT.YAML, cfg.DATA_SPLIT.SUBSET, output=cfg.OUTPUT)
        hdf5_detections, hdf5_groundtruth, soc_data = load_evaulation_files(
            cfg.DATA.PATH, dataset=dataset, max_workers=cfg.DATA.MAX_THREADS
        )

        supported_dataset = support_dataset_with_file_paths(hdf5_groundtruth, hdf5_detections, soc_data)
        evaluator = Metrics(supported_dataset, cfg.METRICS, cfg.OUTPUT)
        evaluator.evaluate()


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-

import unittest

from uval.utils.label_naming import label_long_to_short, label_short_to_long

example_volume_id_1 = "BAGGAGE_20210901_064516_00000"
example_volume_id_2 = "BAGGAGE_20190719_203712_28467"
example_label_id_1 = "%_label_1"
example_label_id_2 = "%_label_2"
example_invalid_label_id = "X_label_2"

example_long_label_name_1 = "BAGGAGE_20210901_064516_00000_label_1"
example_long_label_name_2 = "BAGGAGE_20190719_203712_28467_label_2"


class TestNamingMethods(unittest.TestCase):
    def test_label_short_to_long(self):
        msg = "The long label name does not match!"
        self.assertEqual(label_short_to_long(example_volume_id_1, example_label_id_1), example_long_label_name_1, msg)
        self.assertEqual(label_short_to_long(example_volume_id_2, example_label_id_2), example_long_label_name_2, msg)

    def test_label_long_to_short(self):
        msg = "The short label name does not match!"
        msg_char = "Short id must contain '%' character!"

        short_id_1 = label_long_to_short(example_volume_id_1, example_long_label_name_1)
        self.assertEqual(short_id_1, example_label_id_1, msg)
        self.assertTrue("%" in short_id_1, msg_char)

        short_id_2 = label_long_to_short(example_volume_id_2, example_long_label_name_2)
        self.assertEqual(short_id_2, example_label_id_2, msg)
        self.assertTrue("%" in short_id_1, msg_char)

    def test_label_short_to_long_throws_exception(self):
        self.assertRaises(ValueError, label_short_to_long, example_volume_id_1, example_invalid_label_id)

    def test_label_long_to_short_throws_exception(self):
        self.assertRaises(ValueError, label_long_to_short, example_volume_id_2, example_long_label_name_1)


if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-

import sys
import unittest
from os import path, unlink
from pprint import pprint

import numpy as np

from uval.tests.test_hdf5 import write_sample_h5_file
from uval.utils.hdf5_io import UvalHdfFileOutput
from uval.utils.hdf5_verification import verify_hdf5_files, verify_single_hdf5_file

# to let the unittest find the files using the relative path
current_dir = path.dirname(__file__)
sys.path.insert(0, current_dir)


class TestVerificationMethods(unittest.TestCase):
    def test_verify_multiple_h5_files(self):
        sample_files = [
            f"{current_dir}/../../../data/hdf5/detections/BAGGAGE_20171205_072023_012345.det.h5",
            f"{current_dir}/../../../data/hdf5/raw/BAGGAGE_20180320_080245_124166.gt.h5",
        ]
        for h5_file in sample_files:
            print(f"\nChecking {h5_file}:")
            problems = verify_single_hdf5_file(h5_file)
            self.assertEqual(len(problems), 0)

    def test_verify_sample_h5_files(self):
        problems_det = verify_hdf5_files(f"{current_dir}/../../../data/hdf5/detections")
        self.assertEqual(len(problems_det), 0)
        problems_raw = verify_hdf5_files(f"{current_dir}/../../../data/hdf5/raw")
        self.assertEqual(len(problems_raw), 0)

    def test_create_and_verify_voldata_h5_file(self):
        # Generating voldata file and checking it:
        test_file = "test.voldata.h5"
        with UvalHdfFileOutput(test_file) as outf:
            outf.file_meta = {"det_version": "20-99-05"}
            outf.volume_meta = {
                "id": "BAG",
                "file_md5": "1a3ef6",
                "full_shape": np.array([1, 3, 2]),
                "is_cropped": 1,
                "roi_start": np.array([0, 0, 0]),
                "roi_shape": np.array([620, 420, 300]),
            }
            outf.volume = np.zeros((1, 3, 2), dtype=np.uint16)
        print("\nChecking generated volcache file:")
        problems = verify_single_hdf5_file(test_file)
        self.assertEqual(len(problems), 0)
        unlink(test_file)

    def test_create_and_verify_full_h5_file(self):
        sample_name = "test.det.h5"
        self.assertTrue(write_sample_h5_file(sample_name), f"Could not create sample h5 file: {sample_name}")
        problems = verify_single_hdf5_file(sample_name)
        self.assertEqual(len(problems), 0)
        unlink(sample_name)

    def test_create_and_verify_h5_file_with_invalid_name(self):
        invalid_name = "test.h5"
        self.assertTrue(write_sample_h5_file(invalid_name), f"Could not create sample h5 file: {invalid_name}")
        problems = verify_single_hdf5_file(invalid_name)
        self.assertEqual(len(problems), 1)
        pprint(problems)
        unlink(invalid_name)


if __name__ == "__main__":
    unittest.main()



from uval.config.config_utils import default_argument_parser, setup_from_args  # type: ignore
from uval.context import get_context  # type: ignore
from uval.stages import load_evaulation_files, support_dataset_with_file_paths  # type: ignore
from uval.stages.dataset_specification import load_datasplit  # type: ignore
from uval.stages.metrics import Metrics  # type: ignore

__all__ = [
    "setup_from_args",
    "default_argument_parser",
    "get_context",
    "load_evaulation_files",
    "support_dataset_with_file_paths",
    "load_datasplit",
    "Metrics",
]



# -*- coding: utf-8 -*-

import importlib.util
import unittest

from uval.utils.log import RootLogger


class TestLogMethods(unittest.TestCase):
    def test_root_logger_init(self):
        root_logger = RootLogger()
        self.assertEqual(len(root_logger.root_logger.handlers), 1)
        self.assertEqual(root_logger.root_logger.level, 10)

    @unittest.skip("Not used for now")
    def test_load_uval_script(self):
        spec = importlib.util.spec_from_file_location("a_b", r"../stages/stage.py")
        mod = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(mod)
        self.assertEqual(mod.__name__, "a_b")
        self.assertEqual(mod.__sizeof__(), 56)


if __name__ == "__main__":
    unittest.main()
